/*
 * Class for network communication between game server and client.
 * 
 * @author Frode Voll Aasen
 * @author Tom Christian Bjrlo Johannesen
 * 
 */
import java.net.*;
import java.io.*;
import java.util.ArrayList;


public class ClientConnection extends Thread {
	private GameServer gs;
	private Socket client;
	private String name;
	private ArrayList<SDLTimer> timers = new ArrayList<SDLTimer>(10);
	private ArrayList<Signal> input = new ArrayList<Signal>(100);
	
	/*
	 * Class constuctor
	 * 
	 * @param gs	Reference to GameServer object
	 * @param client	Connection to client
	 */
	public ClientConnection(GameServer gs, Socket client) {
		this.gs = gs;
		this.client = client;
	}
	
	/*
	 * Returns the IP address of the client
	 * 
	 * @return Client IP
	 */
	public String getClientIP() {
		return client.getInetAddress().getHostName();
	}
	
	/*
	 * Class for handling the reading and writing over the network
	 * 
	 */
	private class socket extends Thread {
		private BufferedReader network;
		public boolean running = true;
		private ClientConnection parent;
		private Socket client;
		
		public socket(Socket client, ClientConnection parent) {
			try {
				network = new BufferedReader(new InputStreamReader(client.getInputStream()));
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			this.client = client;
			this.parent = parent;
		}
		
		public void run() {
			while(running) {
				try {
					//if(network.ready()) {
						String network_in;
						network_in = network.readLine();
						if(network_in != null)
							network_in.trim();
						if(!network_in.equals("")) {
							String[] signal_in = network_in.split(";");
							ArrayList netInParam = new ArrayList();
							for(int i = 1; i < signal_in.length; i++) {
								netInParam.add(signal_in[i].trim());
							}
							parent.addInput(new Signal(signal_in[0].trim(), netInParam, false));
						}
					//}
				} catch (IOException e) {
					// TODO Auto-generated catch block
					//e.printStackTrace();
					running = false;
					try {
						client.close();
					} catch (IOException e2) {
						
					}
					System.err.println("Client disconnected from GS");
					parent.addInput(new Signal("Disconnect", null, true));
				}
			}
		} 
	}
	
	public void run() {
		System.out.println("ClientConnection starting... Connection from " + client.getRemoteSocketAddress().toString());
		String state = "Ready";
		boolean running = true;
		boolean isEmpty = true;
		socket networkthread = this.new socket(client, this);
		networkthread.start();
		
		while(running) {
			while(isEmpty) { // While input queue has no input
				try {
					Thread.sleep(5);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				for(int i = 0; i < timers.size(); i++) {
					if(timers.get(i).isExpired()) {
						this.addInput(timers.remove(i).getSignal()); //TODO should be added with priority
					}
				}
				synchronized(input) { // protect variable
					if(!input.isEmpty()) isEmpty = false;
				}
			}
			isEmpty = true;
			synchronized(input) {
				Signal s = input.remove(0);
				String signal = s.name;
				System.out.println("ClientConnection received signal: " + signal + " in state " + state);
				if(state.equals("Ready")) {
					if(signal.equals("LogIn")) {
						s.parameters.add(this);
						gs.db.addInput(s);
						state = "WaitForDB";
					}
				}
				else if(state.equals("WaitForDB")) {
					if(signal.equals("LoginOK")) {
						SDLTimer t = new SDLTimer(300000, new Signal("Timeout", null, true));
						t.start();
						timers.add(t);
						this.signalClient(s);
						state = "LoggedIn";
					}
					if(signal.equals("LoginAccessDenied")) {
						this.signalClient(s);
						state = "Ready";
					}
					if(signal.equals("LogIn")) {
						this.addInput(s); // TODO: possible starvation
					}
				}
				else if(state.equals("LoggedIn")) {
					if(signal.equals("Timeout")) {
						System.out.println("Timeout");
						running = false;
					}
					if(signal.equals("ConnectGS")) {
						this.name = (String) s.parameters.get(0);
						s.parameters.add(this);
						this.gs.zm.addInput(s);
						state = "Waiting";
					}
				}
				else if(state.equals("Waiting")) {
					if(signal.equals("ConnectGSOK")) {
						this.signalClient(s);
						state = "Connected";
					}
				}
				else if(state.equals("Connected")) {
					if(signal.equals("Disconnect")) {
						running = false;
						ArrayList param = new ArrayList();
						param.add(this.name);
						this.gs.zm.addInput(new Signal("Disconnect", param, true));
					}
					if(signal.equals("ManageZone")) {
						this.signalClient(s);
					}
					if(signal.equals("NewZoneMaster")) {
						SDLTimer t = new SDLTimer(3000, new Signal("SendNewZoneMaster", s.parameters, true));
						t.start();
						timers.add(t);
					}
					if(signal.equals("SendNewZoneMaster")) {
						this.signalClient(new Signal("NewZoneMaster", s.parameters, true));
					}
				}
				else if(state.equals("WaitForClient")) {
					if(signal.equals("Disconnect")) {
						running = false;
						ArrayList param = new ArrayList();
						param.add(this.name);
						this.gs.zm.addInput(new Signal("Disconnect", param, true));
					}
					if(signal.equals("IAmAlive")) {
						// TODO send NewZoneMaster
						state = "Connected";
					}
				}
			}
		}
	}

	/*
	 * Adds a signal to the back of the signal queue.
	 * 
	 * @param input The signal to be added
	 */
	public void addInput(Signal input) {
		synchronized(this.input) {
			this.input.add(input);
		}
	}
	
	/*
	 * Sends a signal over the network
	 * 
	 * @param output The signal to be sent
	 */
	public void signalClient(Signal output) {
		try {
			PrintStream os = new PrintStream(client.getOutputStream());
			String out = output.name;
			if(output.parameters != null)
				for(int i = 0; i < output.parameters.size(); i++) {
					out = out + ";" + output.parameters.get(i).toString();
				}
			os.print(out + "\n");
		} catch (IOException e) {
			System.err.println(e);
		}
	}
}

