/*
 * The Database class contains functionality for retrieving and storing game data.
 * 
 * @author Frode Voll Aasen
 * @author Tom Christian Bj�rlo Johannesen
 */
import java.util.*;

public class Database extends Thread {
	private GameServer gs;
	private String ip;
	private String database;
	private String username;
	private String password;

	private ArrayList<Signal> input = new ArrayList<Signal>(100);

	private HashMap<String, PlayerCharacter> characters = new HashMap<String, PlayerCharacter>();
	
	public void run() {
		System.out.println("Database starting...");
		String state = "Ready";
		boolean running = true;
		boolean noInput = true;
		while(running) {
			while(noInput) {
				try {
					Thread.sleep(1);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				synchronized(input) {
					if(!input.isEmpty()) noInput = false;
				}
			}
			noInput = true;
			synchronized(input) {
				Signal s = input.remove(0);
				String signal = s.name;
				System.out.println("Database received signal: " + signal + " in state " + state);
				if(state.equals("Ready")) {
					if(signal.equals("LogIn")) {
						ClientConnection cc = (ClientConnection) s.parameters.get(2);
						if(this.LogIn((String) s.parameters.get(0), (String) s.parameters.get(1))) {
							cc.addInput(new Signal("LoginOK", null, true));
						} else {
							cc.addInput(new Signal("LoginAccessDenied", null, true));
						}
					}
					if(signal.equals("Backup")) {
						String update = (String) s.parameters.get(0);
						String[] updates = update.split(";");
						for(int i = 0; i < updates.length; i++) {
							String[] change = (updates[i]).split(":");
							PlayerCharacter pc = characters.get(change[0]);
							pc.posX = Double.parseDouble(change[1]);
							pc.posY = Double.parseDouble(change[2]);
							pc.posZ = Double.parseDouble(change[3]);
						}
					}
					if(signal.equals("ObjectAdd")) {
						// TODO add object
					}
					if(signal.equals("GetZoneInfo")) {
						Zone z = (Zone) s.parameters.get(0);
						
						GameObjectList zoneinfo = this.GetZoneInfo(z);
						ArrayList param = new ArrayList();
						param.add(z.ZoneID);
						param.add(zoneinfo);
						gs.zm.addInput(new Signal("ZoneInfo", param, true));
					}
					if(signal.equals("GetObject")) {
						GameObject gameobject = this.GetObject((Integer) s.parameters.get(0));

						//gs.zm.addInput(res);
					}
					if(signal.equals("ConnectGS")) {
						PlayerCharacter pc = this.GetPlayerByName((String) s.parameters.get(0));
						ArrayList param = new ArrayList();
						param.add(pc);
						gs.zm.addInput(new Signal("ConnectGSOK", param, true));
					}
				}
			}
		}
	}
	
	/*
	 * Adds a signal to the back of the signal queue.
	 * 
	 * @param input The signal to be added
	 */
	public void addInput(Signal input) {
		synchronized(this.input) {
			this.input.add(input);
		}
	}

	/*
	 * Class constructor
	 * 
	 * @param gs	Reference to GameServer object
	 * @param ip	Database IP address
	 * @param database	Database name
	 * @param username	Database username
	 * @param password	Database password
	 */
	public Database(GameServer gs, String ip, String database, String username, String password) {
		this.gs = gs;
		this.ip = ip;
		this.database = database;
		this.username = username;
		this.password = password;
		/*
		this.characters.put("Tom", new PlayerCharacter(0, 30, 0, 10, 0, (((double)3/(double)4)*Math.PI), 0, "Tom"));
		this.characters.put("Frode", new PlayerCharacter(1, 25, 0, 40, 0, (((double)7/(double)4)*Math.PI), 0, "Frode"));
		this.characters.put("Player2", new PlayerCharacter(2, 5, 0, 5, 0, (((double)7/(double)4)*Math.PI), 0, "Player2"));
		this.characters.put("Player3", new PlayerCharacter(3, 20, 0, 5, 0, (((double)7/(double)4)*Math.PI), 0, "Player3"));
		this.characters.put("Player4", new PlayerCharacter(4, 40, 0, 5, 0, (((double)7/(double)4)*Math.PI), 0, "Player4"));
		this.characters.put("Player5", new PlayerCharacter(5, 60, 0, 5, 0, (((double)7/(double)4)*Math.PI), 0, "Player5"));
		this.characters.put("Player6", new PlayerCharacter(6, 80, 0, 5, 0, (((double)7/(double)4)*Math.PI), 0, "Player6"));
		this.characters.put("Player7", new PlayerCharacter(7, 20, 0, 80, 0, (((double)7/(double)4)*Math.PI), 0, "Player7"));
		this.characters.put("Player8", new PlayerCharacter(8, 40, 0, 80, 0, (((double)7/(double)4)*Math.PI), 0, "Player8"));
		this.characters.put("Player9", new PlayerCharacter(9, 80, 0, 80, 0, (((double)7/(double)4)*Math.PI), 0, "Player9"));
		this.characters.put("Player10", new PlayerCharacter(10, 60, 0, 60, 0, (((double)7/(double)4)*Math.PI), 0, "Player10"));
		this.characters.put("Player11", new PlayerCharacter(11, 90, 0, 10, 0, (((double)7/(double)4)*Math.PI), 0, "Player11"));
		this.characters.put("Player12", new PlayerCharacter(12, 80, 0, 30, 0, (((double)7/(double)4)*Math.PI), 0, "Player12"));
		this.characters.put("Player13", new PlayerCharacter(13, 5, 0, 80, 0, (((double)7/(double)4)*Math.PI), 0, "Player13"));
		this.characters.put("Player14", new PlayerCharacter(14, 25, 0, 75, 0, (((double)7/(double)4)*Math.PI), 0, "Player14"));
		this.characters.put("Player15", new PlayerCharacter(15, 60, 0, 40, 0, (((double)7/(double)4)*Math.PI), 0, "Player15"));
		this.characters.put("Player16", new PlayerCharacter(16, 90, 0, 45, 0, (((double)7/(double)4)*Math.PI), 0, "Player16"));
		this.characters.put("Player17", new PlayerCharacter(17, 45, 0, 90, 0, (((double)7/(double)4)*Math.PI), 0, "Player17"));
		this.characters.put("Player18", new PlayerCharacter(18, 10, 0, 55, 0, (((double)7/(double)4)*Math.PI), 0, "Player18"));
		this.characters.put("Player19", new PlayerCharacter(19, 98, 0, 98, 0, (((double)7/(double)4)*Math.PI), 0, "Player19"));
		*/
		
		this.characters.put("Tom", new PlayerCharacter(0, 25, 0, 60, 0, (((double)3/(double)4)*Math.PI), 0, "Tom"));
		this.characters.put("Frode", new PlayerCharacter(1, 30, 0, 60, 0, (((double)7/(double)4)*Math.PI), 0, "Frode"));
		this.characters.put("Player2", new PlayerCharacter(2, 30, 0, 60, 0, (((double)7/(double)4)*Math.PI), 0, "Player2"));
		this.characters.put("Player3", new PlayerCharacter(3, 30, 0, 60, 0, (((double)7/(double)4)*Math.PI), 0, "Player3"));
		this.characters.put("Player4", new PlayerCharacter(4, 30, 0, 60, 0, (((double)7/(double)4)*Math.PI), 0, "Player4"));
		this.characters.put("Player5", new PlayerCharacter(5, 30, 0, 60, 0, (((double)7/(double)4)*Math.PI), 0, "Player5"));
		this.characters.put("Player6", new PlayerCharacter(6, 30, 0, 60, 0, (((double)7/(double)4)*Math.PI), 0, "Player6"));
		this.characters.put("Player7", new PlayerCharacter(7, 30, 0, 60, 0, (((double)7/(double)4)*Math.PI), 0, "Player7"));
		this.characters.put("Player8", new PlayerCharacter(8, 30, 0, 60, 0, (((double)7/(double)4)*Math.PI), 0, "Player8"));
		this.characters.put("Player9", new PlayerCharacter(9, 30, 0, 60, 0, (((double)7/(double)4)*Math.PI), 0, "Player9"));
		this.characters.put("Player10", new PlayerCharacter(10, 30, 0, 60, 0, (((double)7/(double)4)*Math.PI), 0, "Player10"));
		this.characters.put("Player11", new PlayerCharacter(11, 30, 0, 60, 0, (((double)7/(double)4)*Math.PI), 0, "Player11"));
		this.characters.put("Player12", new PlayerCharacter(12, 30, 0, 60, 0, (((double)7/(double)4)*Math.PI), 0, "Player12"));
		this.characters.put("Player13", new PlayerCharacter(13, 30, 0, 60, 0, (((double)7/(double)4)*Math.PI), 0, "Player13"));
		this.characters.put("Player14", new PlayerCharacter(14, 30, 0, 60, 0, (((double)7/(double)4)*Math.PI), 0, "Player14"));
		this.characters.put("Player15", new PlayerCharacter(15, 30, 0, 60, 0, (((double)7/(double)4)*Math.PI), 0, "Player15"));
		this.characters.put("Player16", new PlayerCharacter(16, 30, 0, 60, 0, (((double)7/(double)4)*Math.PI), 0, "Player16"));
		this.characters.put("Player17", new PlayerCharacter(17, 30, 0, 60, 0, (((double)7/(double)4)*Math.PI), 0, "Player17"));
		this.characters.put("Player18", new PlayerCharacter(18, 30, 0, 60, 0, (((double)7/(double)4)*Math.PI), 0, "Player18"));
		this.characters.put("Player19", new PlayerCharacter(19, 30, 0, 60, 0, (((double)7/(double)4)*Math.PI), 0, "Player19"));
		
		this.Connect();
	}

	private boolean Connect() {
		return true;
	}

	/*
	 * Validates a users username and password.
	 * 
	 * @param username	Users username
	 * @param password	Users password
	 * @return 			true if credentials are valid, false if not.
	 */
	private boolean LogIn(String username, String password) {
		// retrieve username & password from database;
		// "SELECT password FROM authentication WHERE username = " + username;
		String truePassword = "tisu";
		if(password.equals(truePassword))
			return true;

		return false;
	}

	private void ObjectAdd(GameObject object) {
		// write object data to database
	}

	/*
	 * Retrieves a GameObject from the database
	 * 
	 * @param objectID	Object ID of the GameObject to return
	 * @return			Requested GameObject
	 */
	private GameObject GetObject(int objectID) {
		// "SELECT * FROM GameObjects WHERE objectID = " + objectID;
		//gameObject = new GameObject(objectID ....... );
		return new GameObject(1, "", 1,1,1,1,1,1);
	}

	/*
	 * Retrieves a PlayerCharacter object
	 * 
	 * @param name	Characters name
	 * @return		Requested PlayerCharacter object
	 */
	private PlayerCharacter GetPlayerByName(String name) {
		PlayerCharacter p = this.characters.get(name);
		return p;
	}
	
	/*
	 * Retrieves a list of GameObjects for a zone
	 *  
	 * @param z		Zone object for the zone to retrieve objects.
	 * @return		GameObjectList object containing objects in the zone.
	 */
	private GameObjectList GetZoneInfo(Zone z) {
		//SELECT * FROM GameObjects WHERE posX >= x && posX < (x+w) && posZ >= z && posZ < (z+h)
		GameObjectList go = new GameObjectList();
		
		return go;
	}

	private void Backup(String[] changes) {
		// UPDATE ....
	}
}