/**
 * The GameClient is the main class of a gameclient and holds the data for the client. 
 * It is a reference point for the other classes involving the Clients action.
 * 
 * @author Frode
 * @author Tom-Christian
 */


import java.awt.Color;
import java.awt.Graphics;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.IOException;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import javax.swing.JFrame;
import javax.swing.JPanel;


public class GameClient {
public final static String GameServerIP = "129.241.209.129";
public final static int	GameServerZMPort = 4493;
public final static int gsport = 4492;
public final static int zmport = 4510;
public final static int worldWidth = 100;
public final static int worldHeight = 100;

private ArrayList<SDLTimer> timers = new ArrayList<SDLTimer>(10);
private ArrayList<Signal> input = new ArrayList<Signal>(100);
private ConcurrentHashMap<String, PlayerCharacter> characters = new ConcurrentHashMap<String, PlayerCharacter>();
private ConcurrentHashMap<Integer, GameObject> objects = new ConcurrentHashMap<Integer, GameObject>();
private ZoneMasterHandler zmh;
private String CharacterName = "Tom";
private int CharacterID = 0;

/**
 * Starts the state machine for the client
 * 
 * @param args
 */
	public static void main(String[] args){
	GameClient gc = new GameClient();
	JFrame window = new JFrame("GameClient");
	window.setVisible(true);
	DrawingPanel panel = gc.new DrawingPanel(gc);
	panel.setFocusTraversalKeysEnabled(false);
	panel.setFocusable(true);
	panel.requestFocusInWindow();
	window.add(panel);
	window.setSize(407, 433);
	window.setResizable(false);

	GameServerHandler gsh = new GameServerHandler(gc);
	gsh.start();		
	
	System.out.println("GameClient starting...");
	ArrayList param = new ArrayList();
	param.add("Frode");
	param.add("tisu");
	gsh.addInput(new Signal("LogIn", param, true));
	
	String state = "LoggingIn";
	boolean running = true;
	boolean isEmpty = true;
	while(running) {
		while(isEmpty) { // While input queue has no input
			for(int i = 0; i < gc.timers.size(); i++) {
				if(gc.timers.get(i).isExpired()) {
					gc.addInput(gc.timers.remove(i).getSignal()); //TODO should be added with priority
				}
			}
			synchronized(gc.input) { // protect variable
				if(!gc.input.isEmpty()) isEmpty = false;
			}
		}
		isEmpty = true;
		synchronized(gc.input) {
			Signal s = gc.input.remove(0);
			String signal = s.name;
			if(!signal.equals("CheckStates"))
			System.out.println("GameClient received signal: " + signal + " in state " + state);
			if(state.equals("Started")) {
				if(signal.equals("LogIn")) {
				}
			}
			else if(state.equals("LoggingIn")){
				if(signal.equals("LoginOK")){
				param = new ArrayList();
				param.add(gc.CharacterName);
				gsh.addInput(new Signal("ConnectGS", param, true));
				state = "GSConnecting";
				
				//TODO Set state = LoggedIn
				}
				if(signal.equals("LoginAccessDenied")){
					state = "Started";
				}
			}
			else if(state.equals("LoggedIn")){
				//TODO ConnectGS from Game
				}
				
			else if(state.equals("GSConnecting")){
				if(signal.equals("ConnectGSOK")){
					String Challenge = (String) s.parameters.get(0);
					String ZoneMasterIP = (String) s.parameters.get(1);
					int port = Integer.parseInt((String) s.parameters.get(2));
					System.out.println("ConnectGS contains challenge: " + Challenge + " from ip: " + ZoneMasterIP);
					gc.zmh = new ZoneMasterHandler(gc, ZoneMasterIP, port);
					gc.zmh.start();
					
					param = new ArrayList();
					param.add(Challenge);
					param.add(ZoneMasterIP);
					param.add(gc.CharacterName);
					gc.zmh.addInput(new Signal("ConnectZM", param, true));
					 state = "ZMConnecting";
				}
			}
			else if(state.equals("ZMConnecting")){
				if(signal.equals("ConnectZMOK")){
					System.out.println("Should forward the signal to the game");
					SDLTimer t = new SDLTimer(2000, new Signal("CheckStates", null, true));
					t.start();
					gc.timers.add(t);
					state = "Connected";
				}
				if(signal.equals("ConnectZMFail")){
				/*String error = (String) s.parameters.get(0);
				System.out.println(error);*/
				
				gsh.addInput(s);
				state = "LoggedIn";
				}
			}
			else if(state.equals("Connected")){
				if(signal.equals("NewZoneMaster")){
						gc.zmh.addInput(s);
						String ZoneMasterIP = (String) s.parameters.get(0);
						String Challenge = (String) s.parameters.get(1);
						int port = Integer.parseInt((String) s.parameters.get(2));
						ZoneMasterHandler zmh = new ZoneMasterHandler(gc, ZoneMasterIP, port);
						zmh.start();
						gc.zmh = zmh;
						
						
						param = new ArrayList();
						param.add(Challenge);
						param.add(ZoneMasterIP);
						param.add(gc.CharacterName);
						zmh.addInput(new Signal("ConnectZM", param, true));
						
						state = "ZMConnecting";
				}
				if(signal.equals("ChangeState")){
					if(s.internal){
						synchronized (gc.characters) {
							
							PlayerCharacter o = (PlayerCharacter) gc.objects.get(Integer.parseInt((String) s.parameters.get(0)));
							if(o != null) {
								try {
									Field field = o.getClass().getField((String) s.parameters.get(1));
									String type = field.getType().toString();
									if(type.equals("boolean")) {
										field.setBoolean(o, Boolean.valueOf( (String) s.parameters.get(2) ));
									}
									if(type.equals("double")) {
										field.setDouble(o, Double.valueOf( (String) s.parameters.get(2) ));
									}
								} catch (SecurityException e) {
									// TODO Auto-generated catch block
									e.printStackTrace();
								} catch (NoSuchFieldException e) {
									// TODO Auto-generated catch block
									e.printStackTrace();
								} catch (IllegalArgumentException e) {
									// TODO Auto-generated catch block
									e.printStackTrace();
								} catch (IllegalAccessException e) {
									// TODO Auto-generated catch block
									e.printStackTrace();
								}
								if(o.moving) {
									System.out.println(o.name + " is now moving");
								} else {
									System.out.println(o.name + " is now standing still");
								}	
							}
						}
					}
					else{
						gc.zmh.addInput(s);
						state = "Connected";
					}
				}
				if(signal.equals("CallProcedure")){
					if(s.internal){
						//TODO Send CallProcedure to game
					}
					else{
						gc.zmh.addInput(s);
						state = "Connected";
					}
				}
				if(signal.equals("ObjectAdd")){
					if(s.internal){
						synchronized(gc.characters) {
							if(((GameObject) ((GameObjectList) s.parameters.get(0)).GameObjects.get(0)).type.equals("PlayerCharacter")) {
								String name = ((PlayerCharacter) ((GameObjectList) s.parameters.get(0)).GameObjects.get(0)).name;
								PlayerCharacter pc = (PlayerCharacter) ((GameObjectList) s.parameters.get(0)).GameObjects.get(0);
								gc.characters.put(name, pc);
								gc.objects.put(pc.objectID, pc);
								//ZMClientConnection c = clients.get(name);
							}
						}
						//objects.put(((GameObject) ((GameObjectList) s.parameters.get(0)).GameObjects.get(0)).objectID, (GameObject) ((GameObjectList) s.parameters.get(0)).GameObjects.get(0));	
					}
					else{
						gc.zmh.addInput(s);
						System.out.println(s.parameters.get(0).toString());
						state = "Connected";
					}
					
				}
				if(signal.equals("CheckStates")) {
					panel.requestFocusInWindow();
					synchronized(gc.characters) {
					Set<String> keys = gc.characters.keySet();
					
					Iterator<String> i = keys.iterator();
					while(i.hasNext()) {
						String key = i.next();
						PlayerCharacter pc = gc.characters.get(key);
						Date now = new Date();
						long currentTime = now.getTime();
						if(pc.moving) {
							if(pc.lastStateCheck > 0) {
								long deltaTime = currentTime - pc.lastStateCheck;
								
								pc.posX += ((double)deltaTime/1000)*(1)*Math.sin(pc.rotY);
								pc.posZ += ((double)deltaTime/1000)*(-1)*Math.cos(pc.rotY);
							}	
						}
						pc.lastStateCheck = currentTime;
					}
					
					SDLTimer tt = new SDLTimer(20, new Signal("CheckStates", null, true));
					tt.start();
					gc.timers.add(tt);
					panel.repaint();
					panel.requestFocusInWindow();
					gc.removePlayersOutsideAOI();
					}
				}
				if(signal.equals("AOIUpdate")){
					synchronized(gc.characters) {
					GameObjectList gol = (GameObjectList) s.parameters.get(0);
					for(int i = 0; i < gol.GameObjects.size(); i++) {
						GameObject go = gol.GameObjects.get(i);
						if(go.type.equals("PlayerCharacter")) {
							PlayerCharacter p = gc.characters.get(((PlayerCharacter) go).name);
							if(p != null && !p.equals(go)) {
								//gc.characters.get(((PlayerCharacter) go).name).moving = ((PlayerCharacter) go).moving;
								gc.characters.get(((PlayerCharacter) go).name).posX = ((PlayerCharacter) go).posX;
								gc.characters.get(((PlayerCharacter) go).name).posY = ((PlayerCharacter) go).posY;
								gc.characters.get(((PlayerCharacter) go).name).posZ = ((PlayerCharacter) go).posZ;
								gc.characters.get(((PlayerCharacter) go).name).rotX = ((PlayerCharacter) go).rotX;
								gc.characters.get(((PlayerCharacter) go).name).rotY = ((PlayerCharacter) go).rotY;
								gc.characters.get(((PlayerCharacter) go).name).rotZ = ((PlayerCharacter) go).rotZ;
							}
						}
					}
					}
				}
				
				if(signal.equals("ManageZone")){
					
				GameServerConnection gsm = new GameServerConnection(GameClient.GameServerIP, GameServerZMPort, (Zone) s.parameters.get(0), 4510 +((Zone) s.parameters.get(0)).ZoneID);
				gsm.start();
				}
			}
			}
		}
	}

	/**
	 * Adds a signal from the external GameServer to an internal queue.
	 * 
	 * @param input
	 * The signal that will be added to the input queue.
	 */
public void addInput(Signal input) {
	synchronized(this.input) {
		this.input.add(input);
	}
	

}
/**
 * Remove players that is outside Area Of Interst for player.
 */
private void removePlayersOutsideAOI() {
	synchronized(characters) {
		Set<String> keys = this.characters.keySet();
		
		GameObject p = (GameObject) characters.get(CharacterName);
		
		Iterator<String> i = keys.iterator();
		while(i.hasNext()) {
			String key = i.next();
			PlayerCharacter pc = this.characters.get(key);
			double x = Math.abs((double)pc.posX - (double)p.posX);
			double y = Math.abs((double)pc.posY - (double)p.posY);
			double z = Math.abs((double)pc.posZ - (double)p.posZ);
			double d = Math.sqrt(x*x + y*y + z*z);
			
			if(d > 16) this.characters.remove(key);
		}
	}
}
/**
 * Draws the action performed by the player in the GameWorld panel.
 *
 */
public class DrawingPanel extends JPanel implements KeyListener{
	GameClient gc;
	boolean up = false;
	boolean down = false;
	boolean left = false;
	boolean right = false;
	boolean moving = false;
	
	/**
	 * Draws the GameWorld panel for the GameClient.
	 * @param gc
	 * The GameClient, gc, that the panel is drawn for.
	 */
	DrawingPanel (GameClient gc) {
	    setBackground (Color.WHITE);
	    this.gc = gc;
	    this.addKeyListener(this);
	}
	
	public void paintComponent(Graphics g)   {
	   super.paintComponent(g) ;
	   g.setColor (Color.BLACK);
	   int dy = getSize ().height;
	   int dx = getSize ().width;
	   synchronized(gc.characters) {
		   Set<String> keys = gc.characters.keySet();
			
			Iterator<String> i = keys.iterator();
			while(i.hasNext()) {
				String key = i.next();
				PlayerCharacter pc = gc.characters.get(key);
				
				g.fillOval (((int)pc.posX)*4-2, ((int)pc.posZ)*4-2, 4, 4);
			}
	   }
	}

	@Override
	public void keyPressed(KeyEvent arg0) {
		double rot = 0;
		
		if(arg0.getKeyCode() < 37 || arg0.getKeyCode() > 40) return;
		
		// TODO Auto-generated method stub
		if(arg0.getKeyCode() == 37) {
			if(left) return;
			left = true;
		}
		if(arg0.getKeyCode() == 38) {
			if(up) return;
			up = true;
		}
		if(arg0.getKeyCode() == 39) {
			if(right) return;
			right = true;
		}
		if(arg0.getKeyCode() == 40) {
			if(down) return;
			down = true;
		}


		if(left && !up && !right && !down)
			rot = ((double) 3/ (double) 2) * Math.PI;
		if(!left && up && !right && !down)	
			rot = 0;
		if(!left && !up && right && !down)
			rot = ((double) 1/ (double) 2) * Math.PI;
		if(!left && !up && !right && down)
			rot = Math.PI;
			System.out.println(rot);
		
		synchronized(gc.characters) {
			ArrayList param = new ArrayList();
			param.add(gc.CharacterID);
			param.add("rotY");
			param.add(rot);
			zmh.addInput(new Signal("ChangeState",param, true));
			gc.characters.get(gc.CharacterName).rotY = rot;
			
			if(!moving) {
				moving = true;
				param = new ArrayList();
				param.add(gc.CharacterID);
				param.add("moving");
				param.add("true");
				zmh.addInput(new Signal("ChangeState",param, true));
				gc.characters.get(gc.CharacterName).moving = true;
			}
		}
	}

	@Override
	public void keyReleased(KeyEvent arg0) {
		if(arg0.getKeyCode() < 37 || arg0.getKeyCode() > 40) return;
		// TODO Auto-generated method stub
		if(arg0.getKeyCode() == 37)
			left = false;
		if(arg0.getKeyCode() == 38)
			up = false;
		if(arg0.getKeyCode() == 39)
			right = false;
		if(arg0.getKeyCode() == 40)
			down = false;
		
		
		if(!left && !up && !right && !down) {
			synchronized(gc.characters) {
				moving = false;
				ArrayList param = new ArrayList();
				param.add(gc.CharacterID);
				param.add("moving");
				param.add("false");
				zmh.addInput(new Signal("ChangeState",param, true));
				gc.characters.get(gc.CharacterName).moving = false;
			}
		}
	}

	@Override
	public void keyTyped(KeyEvent arg0) {
		// TODO Auto-generated method stub
		
	}
}



}

