/*
 * Zone Master main class. Contains reference to client connection handler,
 * handles communication with game server and reference to the Zone Model.
 * 
 * @author Frode Voll Aasen
 * @author Tom Christian Bjrlo Johannesen
 * 
 */
import java.io.*;
import java.net.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import com.sun.org.apache.xerces.internal.parsers.DOMParser;

public class GameServerConnection extends Thread {
	private int port;
	
	private ArrayList<SDLTimer> timers = new ArrayList<SDLTimer>(10);
	private ArrayList<Signal> input = new ArrayList<Signal>(100);
	private Socket gameserver;
	private Zone Zone;
	public ZoneModel zm;
	private ClientConnectionHandler cch;
	private HashMap<String, ZMClientConnection> clients = new HashMap<String, ZMClientConnection>();
	private HashMap<String, String> answers = new HashMap<String, String>();
	public boolean running = true;
	/*
	 * Class constructor
	 * 
	 * @param gameServerIP		IP address of the Game Server
	 * @param gameServerPort	Game Server listening port
	 * @param Zone				Zone object of zone to manage
	 * @param port				Client connection handler listening port
	 */
	public GameServerConnection(String gameServerIP, int gameServerPort, Zone Zone, int port) {
		try {
			gameserver = new Socket(gameServerIP, gameServerPort);
		} catch (IOException e) {
			System.err.println(e);
		}
		this.Zone = Zone;
		this.port = port;
	}
	
	public void run() {
		System.out.println("GameServerConnection starting...");
		try {
			Thread.sleep(2000);
		} catch (InterruptedException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		ArrayList param = new ArrayList();
		param.add(this.Zone.ZoneID);
		this.signalGS(new Signal("ZMConnectGS", param, false));
		
		String state = "Idle";
		boolean isEmpty = true;
		try {
			BufferedReader network = new BufferedReader(new InputStreamReader(gameserver.getInputStream()));

			while(running) {
				while(isEmpty) { // While input queue has no input
					Thread.sleep(1);
					if(network.ready()) {
						String network_in = network.readLine();
						network_in.trim();
						if(!network_in.equals("")) {
							String[] signal_in = network_in.split(";");
							ArrayList netInParam = new ArrayList();
							for(int i = 1; i < signal_in.length; i++) {
								netInParam.add(signal_in[i].trim());
							}
							this.addInput(new Signal(signal_in[0].trim(), netInParam, false));
						}
					}
					for(int i = 0; i < timers.size(); i++) {
						if(timers.get(i).isExpired()) {
							this.addInput(timers.remove(i).getSignal()); //TODO should be added with priority
						}
					}
					synchronized(input) { // protect variable
						if(!input.isEmpty()) isEmpty = false;
					}
				}
				isEmpty = true;
				synchronized(input) {
					Signal s = input.remove(0);
					String signal = s.name;
					System.out.println("GameServerConnection received signal: " + signal + " in state " + state);
					if(state.equals("Idle")) {
						if(signal.equals("ZoneInfo")) {
							String objects = (String)s.parameters.get(1);
							GameObjectList gol = parseXMLString(objects);
							this.zm = new ZoneModel(Zone, gol, this);
							this.zm.start();
							
							this.cch = new ClientConnectionHandler(this, this.port);
							this.cch.start();
							
							state = "Started";
						}
					}
					else if(state.equals("Started")) {
						if(signal.equals("GetChallengeAnswer")) {
							ZMClientConnection zmc = (ZMClientConnection) s.parameters.get(1);
							param = new ArrayList();
							if(answers.get((String) s.parameters.get(0)) != null) {
								param.add(answers.get((String) s.parameters.get(0)));
								zmc.addInput(new Signal("ChallengeAnswer", param, true));
							}
						}
						if(signal.equals("ChallengeAnswer")) {
							this.answers.put((String) s.parameters.get(2), (String) s.parameters.get(0));
						}
						if(signal.equals("StopManagingZone")) {
							this.cch.running = false;
							this.zm.td.running = false;
							this.zm.running = false;
							
							running = false;
						}
						if(signal.equals("NewAdjacentZoneMaster")) {
							Zone z = parseXMLZone((String) s.parameters.get(0));
							param = new ArrayList();
							param.add(z);
							param.add(s.parameters.get(1));
							s.parameters = param;
							zm.addInput(s);
						}
						if(signal.equals("Backup")) {
							signalGS(s);
						}
						if(signal.equals("ClientConnected")) {
							this.signalGS(s);
						}
						if(signal.equals("ClientDisconnect")) {
							zm.addInput(s);
						}
						if(signal.equals("ClientTimeout")) {
							this.clients.remove((String) s.parameters.get(0));
							signalGS(s);
						}
						if(signal.equals("ObjectAdd")) {
							GameObjectList go = parseXMLString((String) s.parameters.get(0));
							param = new ArrayList();
							param.add(go);
							zm.addInput(new Signal("ObjectAdd", param, true));
						}
						if(signal.equals("ClientHasNewMaster")) {
							signalGS(s);
						}
					}
				}
			}
			System.out.println("GameServerConnection stopping...");
			zm.running = false;
			cch.running = false;
			cch.server.close();
			/*Set<String> clients = this.clients.keySet();
			Iterator<String> clIter = clients.iterator();
			while(clIter.hasNext()) {
				String cnm = clIter.next();
				(this.clients.get(cnm)).running = false;
			}*/
		} catch (IOException e) {
			System.err.println(e);
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	/*
	 * Adds a signal to the back of the signal queue.
	 * 
	 * @param input The signal to be added
	 */
	public void addInput(Signal input) {
		synchronized(this.input) {
			this.input.add(input);
		}
	}
	
	/*
	 * Sends a signal over the network
	 * 
	 * @param output The signal to be sent
	 */
	public void signalGS(Signal output) {
		try {
			PrintStream os = new PrintStream(gameserver.getOutputStream());
			String out = output.name;
			if(output.parameters != null)
				for(int i = 0; i < output.parameters.size(); i++) {
					out = out + ";" + output.parameters.get(i).toString();
				}
			os.print(out + "\n");
		} catch (IOException e) {
			System.err.println(e);
		}
	}
	
	/*
	 * Parses an XML string to a Zone object
	 * 
	 * @param xml	XML string
	 * @return		Zone object
	 */
	public Zone parseXMLZone(String xml) {
		
		try {
			DOMParser parser = new DOMParser();
			parser.parse(new InputSource(new java.io.StringReader(xml)));
			Document doc = parser.getDocument();

			doc.getDocumentElement().normalize();
		
			NodeList nodeLst = doc.getElementsByTagName("Zone");
			for (int s = 0; s < nodeLst.getLength(); s++) {
				Node fstNode = nodeLst.item(s);
				if (fstNode.getNodeType() == Node.ELEMENT_NODE) {			  
					Element fstElmnt = (Element) fstNode;
					
					NodeList idElmntLst = fstElmnt.getElementsByTagName("ZoneID");
					Element idElmnt = (Element) idElmntLst.item(0);
					NodeList id = idElmnt.getChildNodes();
					int id_ = Integer.parseInt(((Node) id.item(0)).getNodeValue());
					
					NodeList XElmntLst = fstElmnt.getElementsByTagName("x");
					Element XElmnt = (Element) XElmntLst.item(0);
					NodeList X = XElmnt.getChildNodes();
					int X_ = Integer.parseInt(((Node) X.item(0)).getNodeValue());
					
					NodeList YElmntLst = fstElmnt.getElementsByTagName("y");
					Element YElmnt = (Element) YElmntLst.item(0);
					NodeList Y = YElmnt.getChildNodes();
					int Y_ = Integer.parseInt(((Node) Y.item(0)).getNodeValue());
					
					NodeList WElmntLst = fstElmnt.getElementsByTagName("w");
					Element WElmnt = (Element) WElmntLst.item(0);
					NodeList W = WElmnt.getChildNodes();
					int W_ = Integer.parseInt(((Node) W.item(0)).getNodeValue());
					
					NodeList HElmntLst = fstElmnt.getElementsByTagName("h");
					Element HElmnt = (Element) HElmntLst.item(0);
					NodeList H = HElmnt.getChildNodes();
					int H_ = Integer.parseInt(((Node) H.item(0)).getNodeValue());
					
					Zone z = new Zone(id_, X_, Y_, W_, H_);
					return z;
				}
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}
	
	/*
	 * Parses an XML string to a list of GameObjects
	 * 
	 * @param xml	XML string
	 * @return		GameObjectList containing the objects
	 */
	public GameObjectList parseXMLString(String xml) {
		GameObjectList gol = new GameObjectList();
		
		try {
			DOMParser parser = new DOMParser();
			parser.parse(new InputSource(new java.io.StringReader(xml)));
			Document doc = parser.getDocument();

			doc.getDocumentElement().normalize();
			
			NodeList nodeLst = doc.getElementsByTagName("GameObject");
			for (int s = 0; s < nodeLst.getLength(); s++) {
				Node fstNode = nodeLst.item(s);
				if (fstNode.getNodeType() == Node.ELEMENT_NODE) {			  
					Element fstElmnt = (Element) fstNode;
					
					NodeList oidElmntLst = fstElmnt.getElementsByTagName("objectID");
					Element oidElmnt = (Element) oidElmntLst.item(0);
					NodeList oid = oidElmnt.getChildNodes();
					int oid_ = Integer.parseInt(((Node) oid.item(0)).getNodeValue());
					
					NodeList typeElmntLst = fstElmnt.getElementsByTagName("type");
					Element typeElmnt = (Element) typeElmntLst.item(0);
					NodeList type = typeElmnt.getChildNodes();
					String type_ = ((Node) type.item(0)).getNodeValue();
					
					NodeList posXElmntLst = fstElmnt.getElementsByTagName("posX");
					Element posXElmnt = (Element) posXElmntLst.item(0);
					NodeList posX = posXElmnt.getChildNodes();
					double posX_ = Double.parseDouble(((Node) posX.item(0)).getNodeValue());
					
					NodeList posYElmntLst = fstElmnt.getElementsByTagName("posY");
					Element posYElmnt = (Element) posYElmntLst.item(0);
					NodeList posY = posYElmnt.getChildNodes();
					double posY_ = Double.parseDouble(((Node) posY.item(0)).getNodeValue());
					
					NodeList posZElmntLst = fstElmnt.getElementsByTagName("posZ");
					Element posZElmnt = (Element) posZElmntLst.item(0);
					NodeList posZ = posZElmnt.getChildNodes();
					double posZ_ = Double.parseDouble(((Node) posZ.item(0)).getNodeValue());
					
					NodeList rotXElmntLst = fstElmnt.getElementsByTagName("rotX");
					Element rotXElmnt = (Element) rotXElmntLst.item(0);
					NodeList rotX = rotXElmnt.getChildNodes();
					double rotX_ = Double.parseDouble(((Node) rotX.item(0)).getNodeValue());
					
					NodeList rotYElmntLst = fstElmnt.getElementsByTagName("rotY");
					Element rotYElmnt = (Element) rotYElmntLst.item(0);
					NodeList rotY = rotYElmnt.getChildNodes();
					double rotY_ = Double.parseDouble(((Node) rotY.item(0)).getNodeValue());
					
					NodeList rotZElmntLst = fstElmnt.getElementsByTagName("rotZ");
					Element rotZElmnt = (Element) rotZElmntLst.item(0);
					NodeList rotZ = rotZElmnt.getChildNodes();
					double rotZ_ = Double.parseDouble(((Node) rotZ.item(0)).getNodeValue());
					
					GameObject go = new GameObject(oid_, type_, posX_, posY_, posZ_, rotX_, rotY_, rotZ_);
					gol.add(go);
				}
			}
			
			nodeLst = doc.getElementsByTagName("PlayerCharacter");
			for (int s = 0; s < nodeLst.getLength(); s++) {
				Node fstNode = nodeLst.item(s);
				if (fstNode.getNodeType() == Node.ELEMENT_NODE) {			  
					Element fstElmnt = (Element) fstNode;
					
					NodeList oidElmntLst = fstElmnt.getElementsByTagName("objectID");
					Element oidElmnt = (Element) oidElmntLst.item(0);
					NodeList oid = oidElmnt.getChildNodes();
					int oid_ = Integer.parseInt(((Node) oid.item(0)).getNodeValue());
					
					NodeList posXElmntLst = fstElmnt.getElementsByTagName("posX");
					Element posXElmnt = (Element) posXElmntLst.item(0);
					NodeList posX = posXElmnt.getChildNodes();
					double posX_ = Double.parseDouble(((Node) posX.item(0)).getNodeValue());
					
					NodeList posYElmntLst = fstElmnt.getElementsByTagName("posY");
					Element posYElmnt = (Element) posYElmntLst.item(0);
					NodeList posY = posYElmnt.getChildNodes();
					double posY_ = Double.parseDouble(((Node) posY.item(0)).getNodeValue());
					
					NodeList posZElmntLst = fstElmnt.getElementsByTagName("posZ");
					Element posZElmnt = (Element) posZElmntLst.item(0);
					NodeList posZ = posZElmnt.getChildNodes();
					double posZ_ = Double.parseDouble(((Node) posZ.item(0)).getNodeValue());
					
					NodeList rotXElmntLst = fstElmnt.getElementsByTagName("rotX");
					Element rotXElmnt = (Element) rotXElmntLst.item(0);
					NodeList rotX = rotXElmnt.getChildNodes();
					double rotX_ = Double.parseDouble(((Node) rotX.item(0)).getNodeValue());
					
					NodeList rotYElmntLst = fstElmnt.getElementsByTagName("rotY");
					Element rotYElmnt = (Element) rotYElmntLst.item(0);
					NodeList rotY = rotYElmnt.getChildNodes();
					double rotY_ = Double.parseDouble(((Node) rotY.item(0)).getNodeValue());
					
					NodeList rotZElmntLst = fstElmnt.getElementsByTagName("rotZ");
					Element rotZElmnt = (Element) rotZElmntLst.item(0);
					NodeList rotZ = rotZElmnt.getChildNodes();
					double rotZ_ = Double.parseDouble(((Node) rotZ.item(0)).getNodeValue());
					
					NodeList nameElmntLst = fstElmnt.getElementsByTagName("name");
					Element nameElmnt = (Element) nameElmntLst.item(0);
					NodeList name = nameElmnt.getChildNodes();
					String name_ = ((Node) name.item(0)).getNodeValue();
					
					NodeList movingElmntLst = fstElmnt.getElementsByTagName("moving");
					Element movingElmnt = (Element) movingElmntLst.item(0);
					NodeList moving = movingElmnt.getChildNodes();
					boolean moving_ = Boolean.parseBoolean(((Node) moving.item(0)).getNodeValue());
					
					PlayerCharacter go = new PlayerCharacter(oid_, posX_, posY_, posZ_, rotX_, rotY_, rotZ_, name_);
					go.moving = moving_;
					gol.add(go);
				}
			}
			
		} catch (Exception e) {
			e.printStackTrace();
		}
		return gol;
	}
}
