/**
 *  The GameServerHandler class sets up the link between the GameServer and GameClient, 
 *  and forwards signal between the two actors over the network.
 *  
 *  @author Frode
 *  @author Tom-Christian
 */


import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.net.Socket;
import java.util.ArrayList;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

import com.sun.org.apache.xerces.internal.parsers.DOMParser;

public class GameServerHandler extends Thread{
	
	private ArrayList<SDLTimer> timers = new ArrayList<SDLTimer>(10);
	private ArrayList<Signal> input = new ArrayList<Signal>(100);
	private Socket gameserver;
	private GameClient gc;
	
	/**
	 * Constructor that sets up a socket connection based on 
	 * GameServer-IP and GameServer-port.
	 * 
	 * @param g
	 * The GameClient, g, that wants to connect to GameServer.
	 */
	public GameServerHandler(GameClient g){
		this.gc = g;
		try {
			gameserver = new Socket(GameClient.GameServerIP, GameClient.gsport);
		} catch (IOException e) {
			System.err.println(e);
		}
	}
	/**
	 * Starts the state machine, check SDL design
	 */
		public void run() {
			System.out.println("GameServerHandler starting...");
			try {
				Thread.sleep(2000);
			} catch (InterruptedException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
			String state = "Started";
			boolean running = true;
			boolean isEmpty = true;
			try {
				BufferedReader network = new BufferedReader(new InputStreamReader(gameserver.getInputStream()));

				while(running) {
					while(isEmpty) { // While input queue has no input
						Thread.sleep(10);
						if(network.ready()) {
							String network_in = network.readLine();
							network_in.trim();
							if(!network_in.equals("")) {
								String[] signal_in = network_in.split(";");
								ArrayList netInParam = new ArrayList();
								for(int i = 1; i < signal_in.length; i++) {
									netInParam.add(signal_in[i].trim());
								}
								this.addInput(new Signal(signal_in[0].trim(), netInParam, false));
							}
						}
						for(int i = 0; i < timers.size(); i++) {
							if(timers.get(i).isExpired()) {
								this.addInput(timers.remove(i).getSignal()); //TODO should be added with priority
							}
						}
						synchronized(input) { // protect variable
							if(!input.isEmpty()) isEmpty = false;
						}
					}
					isEmpty = true;
					synchronized(input) {
						Signal s = input.remove(0);
						String signal = s.name;
						System.out.println("GameServerHandler received signal: " + signal + " in state " + state);
						if(state.equals("Started")) {
							if(signal.equals("LogIn")) {
								signalGS(s);
								state = "LoggingIn";
							}
						}
						else if(state.equals("LoggingIn")){
							if(signal.equals("LoginOK")){
								gc.addInput(s);
								state = "LoggedIn";
							}
							if(signal.equals("LoginAccesDenied")){
								gc.addInput(s);
								state = "Started";
							}
						}
						else if(state.equals("LoggedIn")){
							if(signal.equals("ConnectGS")){
								signalGS(s);
								state = "GSConnecting";
							}
						}
						else if(state.equals("GSConnecting")){
							if(signal.equals("ConnectGSOK")){
								gc.addInput(s);
								state = "Connected";
							}
							if(signal.equals("ConnectGSFail")){
								gc.addInput(s);
								state = "LoggedIn";
							}
						}
						else if(state.equals("Connected")){
							if(signal.equals("ManageZone")){
								Zone z = parseXMLZone((String) s.parameters.get(0));
								s.parameters.remove(0);
								s.parameters.add(z);
								gc.addInput(s);
							}
							if(signal.equals("NewZoneMaster")){
								gc.addInput(s);
								state = "Connected";
							}
							if(signal.equals("KeepAlive")){
								signalGS(new Signal("IAmAlive", null, false));
								state = "Connected";
							}
							if(signal.equals("ConnectZMFail")){
								signalGS(new Signal("CannotConnectToZM", null, false));
								state = "Connected";
							}
						}
					}
				}
			} catch (IOException e) {
				System.err.println(e);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		/**
		 * Adds a signal from the external GameServer to an internal queue.
		 * 
		 * @param input
		 * The signal that will be added to the input queue.
		 */
		public void addInput(Signal input) {
			synchronized(this.input) {
				this.input.add(input);
			}
		}
		/**
		 * Sends a signal over the network on the OutputStream.
		 * 
		 * @param output
		 * The signal that will be sent over the network.
		 */
		public void signalGS(Signal output) {
			try {
				PrintStream os = new PrintStream(gameserver.getOutputStream());
				String out = output.name;
				if(output.parameters != null)
					for(int i = 0; i < output.parameters.size(); i++) {
						out = out + ";" + output.parameters.get(i).toString();
					}
				os.print(out + "\n");
			} catch (IOException e) {
				System.err.println(e);
			}
		}
		
		/**
		 * Deserialises the XML string to a zone object.
		 * 
		 * @param xml
		 * The xml string that is sent over the network
		 * 
		 * @return
		 * The zone object
		 */
		public Zone parseXMLZone(String xml) {
			
			try {
				DOMParser parser = new DOMParser();
				parser.parse(new InputSource(new java.io.StringReader(xml)));
				Document doc = parser.getDocument();

				doc.getDocumentElement().normalize();
			
				NodeList nodeLst = doc.getElementsByTagName("Zone");
				for (int s = 0; s < nodeLst.getLength(); s++) {
					Node fstNode = nodeLst.item(s);
					if (fstNode.getNodeType() == Node.ELEMENT_NODE) {			  
						Element fstElmnt = (Element) fstNode;
						
						NodeList idElmntLst = fstElmnt.getElementsByTagName("ZoneID");
						Element idElmnt = (Element) idElmntLst.item(0);
						NodeList id = idElmnt.getChildNodes();
						int id_ = Integer.parseInt(((Node) id.item(0)).getNodeValue());
						
						NodeList XElmntLst = fstElmnt.getElementsByTagName("x");
						Element XElmnt = (Element) XElmntLst.item(0);
						NodeList X = XElmnt.getChildNodes();
						int X_ = Integer.parseInt(((Node) X.item(0)).getNodeValue());
						
						NodeList YElmntLst = fstElmnt.getElementsByTagName("y");
						Element YElmnt = (Element) YElmntLst.item(0);
						NodeList Y = YElmnt.getChildNodes();
						int Y_ = Integer.parseInt(((Node) Y.item(0)).getNodeValue());
						
						NodeList WElmntLst = fstElmnt.getElementsByTagName("w");
						Element WElmnt = (Element) WElmntLst.item(0);
						NodeList W = WElmnt.getChildNodes();
						int W_ = Integer.parseInt(((Node) W.item(0)).getNodeValue());
						
						NodeList HElmntLst = fstElmnt.getElementsByTagName("h");
						Element HElmnt = (Element) HElmntLst.item(0);
						NodeList H = HElmnt.getChildNodes();
						int H_ = Integer.parseInt(((Node) H.item(0)).getNodeValue());
						
						Zone z = new Zone(id_, X_, Y_, W_, H_);
						return z;
					}
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
			return null;
		}
		
	}
	

