/*
 * 3D Unit class. Intended to validate movement actions from client.
 * This includes collision detections etc.
 * 
 * @author Frode Voll Aasen
 * @author Tom Christian Bjrlo Johannesen
 */

import java.util.ArrayList;

public class ThreeDUnit extends Thread {
	private ArrayList<SDLTimer> timers = new ArrayList<SDLTimer>(10);
	private ArrayList<Signal> input = new ArrayList<Signal>(100);
	private GameObjectList objects;
	private ZoneModel zm;
	public boolean running = true;
	
	/*
	 * Class constructor
	 * 
	 * @param zm	Reference to ZoneModel
	 * @param objects	List of game objects in zone	
	 */
	public ThreeDUnit(ZoneModel zm, GameObjectList objects) {
		this.zm = zm;
		this.objects = objects;
	}
	
	public void run() {
		System.out.println("3DUnit starting...");
		String state = "Started";
		boolean isEmpty = true;
		try {
			while(running) {
				while(isEmpty) { // While input queue has no input
					Thread.sleep(1);
					for(int i = 0; i < timers.size(); i++) {
						if(timers.get(i).isExpired()) {
							this.addInput(timers.remove(i).getSignal()); //TODO should be added with priority
						}
					}
					synchronized(input) { // protect variable
						if(!input.isEmpty()) isEmpty = false;
					}
				}
				isEmpty = true;
				synchronized(input) {
					Signal s = input.remove(0);
					String signal = s.name;
					System.out.println("3DUnit received signal: " + signal + " in state " + state);
					if(state.equals("Started")) {
						if(signal.equals("ObjectAdd")) {
							// ADD TO MODEL
						}
						if(signal.equals("ChangeState")) {
							// IF VALID:
							zm.addInput(s);
						}
						if(signal.equals("CallProcedure")) {
							// IF VALID:
							zm.addInput(s);
						}
					}
				}
			}
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}
	
	/*
	 * Adds a signal to the back of the signal queue.
	 * 
	 * @param input The signal to be added
	 */
	public void addInput(Signal input) {
		synchronized(this.input) {
			this.input.add(input);
		}
	}
}
