/*
 * Class for network communication between zone master and client.
 * 
 * @author Frode Voll Aasen
 * @author Tom Christian Bjrlo Johannesen
 * 
 */

import java.net.*;
import java.io.*;
import java.util.ArrayList;
import java.util.HashMap;


public class ZMClientConnection extends Thread {
	private GameServerConnection gsc;
	private Socket client;
	private ArrayList<SDLTimer> timers = new ArrayList<SDLTimer>(10);
	private ArrayList<Signal> input = new ArrayList<Signal>(100);
	private int connectRetries = 0;
	public String challengeAnswer;
	private SDLTimer TimeoutTimer;
	private String clientChallenge;
	private String clientCharacterName;
	public boolean running = true;
	
	/*
	 * Class Constructor
	 * 
	 * @param gsc	reference to game server connection
	 * @param client	connection to client
	 */
	public ZMClientConnection(GameServerConnection gsc, Socket client) {
		this.gsc = gsc;
		this.client = client;
	}
	
	/*
	 * Returns remote ip
	 * 
	 * @return remote ip
	 */
	public String getClientIP() {
		return client.getInetAddress().getHostName();
	}
	
	/*
	 * network listening class
	 */
	private class socket extends Thread {
		private BufferedReader network;
		public boolean running = true;
		private ZMClientConnection parent;
		private Socket client;
		
		public socket(Socket client, ZMClientConnection parent) {
			try {
				network = new BufferedReader(new InputStreamReader(client.getInputStream()));
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			this.parent = parent;
			this.client = client;
		}
		
		public void run() {
			while(running) {
				try {
					//if(network.ready()) {
						String network_in;
						network_in = network.readLine();
						if(network_in != null)
							network_in.trim();
						if(!network_in.equals("")) {
							String[] signal_in = network_in.split(";");
							ArrayList netInParam = new ArrayList();
							for(int i = 1; i < signal_in.length; i++) {
								netInParam.add(signal_in[i].trim());
							}
							parent.addInput(new Signal(signal_in[0].trim(), netInParam, false));
						}
					//}
				} catch (IOException e) {
					// TODO Auto-generated catch block
					//e.printStackTrace();
					running = false;
					try {
						client.close();
					} catch (IOException e2) {
						
					}
					parent.running = false;
					System.err.println("Client disconnected from ZM");
					//TODO add disconnect
				}
			}
		} 
	}
	
	public void run() {
		System.out.println("ZMClientConnection starting... Connection from " + client.getRemoteSocketAddress().toString());
		String state = "Ready";
		boolean isEmpty = true;
		socket networkthread = this.new socket(client, this);
		networkthread.start();
		
		try {
			while(running) {
				while(isEmpty) { // While input queue has no input
					for(int i = 0; i < timers.size(); i++) {
						if(timers.get(i).isExpired()) {
							this.addInput(timers.remove(i).getSignal()); //TODO should be added with priority
						}
					}
					synchronized(input) { // protect variable
						if(!input.isEmpty()) isEmpty = false;
					}
					Thread.sleep(5);
				}
				isEmpty = true;
				synchronized(input) {
					Signal s = input.remove(0);
					String signal = s.name;
					System.out.println("ZMClientConnection received signal: " + signal + " in state " + state);
					if(state.equals("Timeout")) {
						running = false;
					}
					if(state.equals("Ready")) {
						if(signal.equals("ConnectZM")) {
							clientChallenge = (String) s.parameters.get(0);
							clientCharacterName = (String) s.parameters.get(2);
							
							ArrayList param = new ArrayList();
							param.add(s.parameters.get(2)); // Character Name
							param.add(this);				// Reference to this connection
							this.gsc.addInput(new Signal("GetChallengeAnswer", param, true));
							
							SDLTimer t = new SDLTimer(5000, new Signal("AnswerTimeout", null, true));
							t.start();
							timers.add(t);
							
							state = "WaitForAnswer";
						}
					}
					else if(state.equals("WaitForAnswer")) {
						if(signal.equals("AnswerTimeout")) {
							signalClient(new Signal("ConnectZMFail", null, false));
						}
						if(signal.equals("ChallengeAnswer")) {
							System.out.println("Client sent challenge: " + clientChallenge);
							System.out.println("Challenge answer is: " + (String) s.parameters.get(0));
							challengeAnswer = (String) s.parameters.get(0);
							if(challengeAnswer != null) {
								if(clientChallenge.equals(challengeAnswer)) {
									signalClient(new Signal("ConnectZMOK", null, false));
									SDLTimer t = new SDLTimer(300000, new Signal("Timeout", null, true));
									t.start();
									timers.add(t);
									TimeoutTimer = t;
									ArrayList param = new ArrayList();
									param.add(clientCharacterName);
									param.add(this);
									
									gsc.zm.addInput(new Signal("ClientConnected", param, true));
									gsc.addInput(new Signal("ClientConnected", param, true));
									
									state = "Connected";
								} else {
									signalClient(new Signal("ConnectZMFail", null, false));
								}
							} else {
								connectRetries++;
								if(connectRetries > 2) {
									running = false;
								} else {
									SDLTimer t = new SDLTimer(2000, s);
									t.start();
									timers.add(t);
								}
							}
						}
					}
					else if(state.equals("Connected")) {
						if(signal.equals("KeepAlive")) {
							TimeoutTimer.reset();
						}
						if(signal.equals("Timeout")) {
							//Timeout disabled
							/*ArrayList param = new ArrayList();
							param.add(this.getClientIP());
							gsc.zm.addInput(new Signal("ClientTimeout", param, true));
							gsc.addInput(new Signal("ClientTimeout", param, true));
							running = false;*/
						}
						if(signal.equals("CallProcedure")) {
							if(s.internal)				// FROM ZoneModel
								signalClient(s);
							else {
								gsc.zm.td.addInput(s);	// FROM Client
								TimeoutTimer.reset();
							}
						}
						if(signal.equals("ChangeState")) {
							if(s.internal)				// FROM ZoneModel
								signalClient(s);
							else {
								gsc.zm.td.addInput(s);	// FROM Client
								TimeoutTimer.reset();
							}
						}
						if(signal.equals("ObjectAdd")) {
							signalClient(s);
						}
						if(signal.equals("AOIUpdate")) {
							signalClient(s);
						}
						if(signal.equals("NewZoneMaster")) {
							signalClient(s);
						}
					}
				}
			}
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/*
	 * Adds a signal to the back of the signal queue.
	 * 
	 * @param input The signal to be added
	 */
	public void addInput(Signal input) {
		synchronized(this.input) {
			this.input.add(input);
		}
	}
	
	/*
	 * Sends a signal over the network
	 * 
	 * @param output The signal to be sent
	 */
	public void signalClient(Signal output) {
		try {
			PrintStream os = new PrintStream(client.getOutputStream());
			String out = output.name;
			if(output.parameters != null)
				for(int i = 0; i < output.parameters.size(); i++) {
					out = out + ";" + output.parameters.get(i).toString();
				}
				System.out.println(out);
			os.print(out + "\n");
		} catch (IOException e) {
			System.err.println(e);
		}
	}
}

