/*
 * Class for network communication between two zone masters.
 * 
 * @author Frode Voll Aasen
 * @author Tom Christian Bjrlo Johannesen
 * 
 */

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.net.Socket;
import java.util.ArrayList;




public class ZMZMConnection extends Thread{
	private ZoneModel zm;
	private Socket otherZM;
	private ArrayList<SDLTimer> timers = new ArrayList<SDLTimer>(10);
	private ArrayList<Signal> input = new ArrayList<Signal>(100);
	public boolean running = true;
	/*
	 * Class constructor
	 * 
	 * @param zm	Reference to Zone Model
	 * @param accept	Connection to other ZM
	 */
	public ZMZMConnection(ZoneModel zm, Socket accept) {
		this.otherZM = accept;
		this.zm = zm;
	}
	
	/*
	 * Returns remote IP address
	 * 
	 * @return IP address
	 */
	public String getRemoteIP() {
		return otherZM.getInetAddress().getHostName();
	}
	
	/*
	 * Class constructor
	 * 
	 * @param zm	Reference to Zone Model
	 * @param remoteIP	IP address of remote ZM
	 * @param remotePort	remote Port
	 */
	public ZMZMConnection(ZoneModel zm, String remoteIP, int remotePort) {
		this.zm = zm;
		try {
			System.out.println("ZMZMConnection connecting to " + remoteIP + ":" + remotePort);
			this.otherZM = new Socket(remoteIP, remotePort);
		} catch (IOException e) {
			System.err.println(e);
		}
	}
	
	/*
	 * Class for listening for data on the network
	 * 
	 */
	private class socket extends Thread {
		private BufferedReader network;
		private ZMZMConnection parent;
		Socket client;
		
		public socket(Socket client, ZMZMConnection parent) {
			try {
				network = new BufferedReader(new InputStreamReader(client.getInputStream()));
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			this.parent = parent;
			this.client = client;
		}
		
		public void run() {
			while(running) {
				try {
					//if(network.ready()) {
						String network_in;
						network_in = network.readLine();
						if(network_in != null)
							network_in.trim();
						if(!network_in.equals("")) {
							String[] signal_in = network_in.split(";");
							ArrayList netInParam = new ArrayList();
							for(int i = 1; i < signal_in.length; i++) {
								netInParam.add(signal_in[i].trim());
							}
							parent.addInput(new Signal(signal_in[0].trim(), netInParam, false));
						}
					//}
				} catch (IOException e) {
					// TODO Auto-generated catch block
					//e.printStackTrace();
					running = false;
					try {
						client.close();
					} catch (IOException e2) {
						
					}
					System.err.println("Other ZM disconnected from ZM");
					parent.addInput(new Signal("Disconnect", null, true));
				}
			}
		} 
	}
	
	public void run() {
		System.out.println("ZMZMConnection starting... Connected to " + otherZM.getRemoteSocketAddress().toString());
		String state = "Ready";
		boolean running = true;
		boolean isEmpty = true;
		socket networkthread = this.new socket(otherZM, this);
		networkthread.start();
		
		while(running) {
			while(isEmpty) { // While input queue has no input
				try {
					Thread.sleep(5);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				for(int i = 0; i < timers.size(); i++) {
					if(timers.get(i).isExpired()) {
						this.addInput(timers.remove(i).getSignal()); //TODO should be added with priority
					}
				}
				synchronized(input) { // protect variable
					if(!input.isEmpty()) isEmpty = false;
				}
			}
			isEmpty = true;
			synchronized(input) {
				Signal s = input.remove(0);
				String signal = s.name;
				System.out.println("ZMZMConnection received signal: " + signal + " in state " + state);
				if(state.equals("Ready")) {
					if(signal.equals("CreateProxy")) {
						if(s.internal) {
							//System.out.println("Internal!");
							signalRemote(s);
						} else {
							//System.out.println("External!");
							zm.addInput(s);
						}
					}
					if(signal.equals("UpdateProxy")) {
						if(s.internal) {
							//System.out.println("Internal!");
							signalRemote(s);
						} else {
							//System.out.println("External!");
							zm.addInput(s);
						}
					}
				}
			}
		}
	}

	/*
	 * Adds a signal to the back of the signal queue.
	 * 
	 * @param input The signal to be added
	 */
	public void addInput(Signal input) {
		synchronized(this.input) {
			this.input.add(input);
		}
	}
	
	/*
	 * Sends a signal over the network
	 * 
	 * @param output The signal to be sent
	 */
	public void signalRemote(Signal output) {
		try {
			PrintStream os = new PrintStream(otherZM.getOutputStream());
			String out = output.name;
			if(output.parameters != null)
				for(int i = 0; i < output.parameters.size(); i++) {
					out = out + ";" + output.parameters.get(i).toString();
				}
			os.print(out + "\n");
		} catch (IOException e) {
			System.err.println(e);
		}
	}
}
