/*
 * Class for managing the game zones.
 * Keeps references to all current players, zonemasters and zones.
 * Decides new zonemasters and manages zone definitions.'
 * 
 * @author Frode Voll Aasen
 * @author Tom Christian Bjrlo Johannesen
 */
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.awt.Color;
import java.awt.Graphics;
import java.io.IOException;
import java.net.*;

import javax.swing.JFrame;
import javax.swing.JPanel;

public class ZoneManager extends Thread {
	private HashMap<String, ClientConnection> clients;			// CharacterName - Connection
	private HashMap<Integer, ZoneMasterConnection> zonemasters;	// ZoneID - ZMConnection
	private HashMap<String, Integer> masterclients;				// CharacterName - ZoneID
	private HashMap<String, PlayerCharacter> currentPCs;		// CharacterName - PlayerCharacter
	protected HashMap<Integer, Zone> zones;						// ZoneID - Zone
	private GameServer gs;
	private static final int numberOfZones = 4;
	private GameServerConnection[] gsZones = new GameServerConnection[ZoneManager.numberOfZones];
	
	/*
	 * Class constructor
	 * 
	 * @param gs	Reference to GameServer
	 */
	public ZoneManager(GameServer gs) {
		this.gs = gs;
		clients = new HashMap<String, ClientConnection>();
		zonemasters = new HashMap<Integer, ZoneMasterConnection>();
		zones = new HashMap<Integer, Zone>();
		currentPCs = new HashMap<String, PlayerCharacter>();
		masterclients = new HashMap<String, Integer>();
	}
	
	private ArrayList<Signal> input = new ArrayList<Signal>(100);

	public void run() {
		System.out.println("ZoneManager starting...");
		
		Zone z = new Zone(0, 0, 0, GameServer.worldWidth/2, GameServer.worldHeight/2);
		zones.put(0, z);
		gsZones[0] = new GameServerConnection(GameServer.localIP, GameServer.zmchPort, z, 4510);
		gsZones[0].start();
		
		Zone z1 = new Zone(1, GameServer.worldWidth/2, 0, GameServer.worldWidth/2, GameServer.worldHeight/2);
		zones.put(1, z1);
		gsZones[1] = new GameServerConnection(GameServer.localIP, GameServer.zmchPort, z1, 4511);
		gsZones[1].start();
		
		Zone z2 = new Zone(2, 0, GameServer.worldHeight/2, GameServer.worldWidth/2, GameServer.worldHeight/2);
		zones.put(2, z2);
		gsZones[2] = new GameServerConnection(GameServer.localIP, GameServer.zmchPort, z2, 4512);
		gsZones[2].start();
		
		Zone z3 = new Zone(3, GameServer.worldWidth/2, GameServer.worldHeight/2, GameServer.worldWidth/2, GameServer.worldHeight/2);
		zones.put(3, z3);
		gsZones[3] = new GameServerConnection(GameServer.localIP, GameServer.zmchPort, z3, 4513);
		gsZones[3].start();
		
		
		JFrame window = new JFrame("GameServer");
		window.setVisible(true);
		DrawingPanel panel = this.new DrawingPanel(this);
		window.add(panel);
		window.setSize(4*GameServer.worldWidth+7, 4*GameServer.worldHeight+33);
		window.setResizable(false);
		
		String state = "Ready";
		boolean running = true;
		boolean isEmpty = true;
		while(running) {
			while(isEmpty) { // While input queue has no input
				try {
					Thread.sleep(50);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				synchronized(input) { // protect variable
					if(!input.isEmpty()) isEmpty = false;
				}
			}
			isEmpty = true;
			synchronized(input) {
				Signal s = input.remove(0);
				String signal = s.name;
				System.out.println("ZoneManager received signal: " + signal + " in state " + state);
				if(state.equals("Ready")) {
					if(signal.equals("ConnectGS")) {
						System.out.println("ZoneManager received 'ConnectGS' from " + (String) s.parameters.get(0));
						this.clients.put((String) s.parameters.get(0), (ClientConnection) s.parameters.get(1));
						gs.db.addInput(s);
					}
					if(signal.equals("ConnectGSOK")) {
						PlayerCharacter pc = (PlayerCharacter) s.parameters.get(0);	// Player's Character
						ZoneMasterConnection zm = zonemasters.get(getZoneMasterByPlayer(pc));		// Connection to ZoneMaster of characters location
						currentPCs.put(pc.name, pc);
						ArrayList param = new ArrayList();
						param.add("Magic");
						param.add(zm.getZoneMasterIP());
						param.add(4510 + getZoneMasterByPlayer(pc));
						ClientConnection c = this.clients.get((String)((PlayerCharacter) s.parameters.get(0)).name);
						c.addInput(new Signal("ConnectGSOK", param, true));			// Send ConnectGSOK(Challenge) to Client

						param = new ArrayList();
						param.add("Magic");
						param.add(c.getClientIP());
						param.add(pc.name);
						zm.addInput(new Signal("ChallengeAnswer", param, true));	// Send ChallengeAnswer(Answer, ClientIP, Player's Character Name) to ZoneMaster
						/*
						param = new ArrayList();
						param.add(pc);
						zm.addInput(new Signal("ObjectAdd", param, true));			// Send ObjectAdd(Player's Character) to ZoneMaster
						*/
					}
					if(signal.equals("GetZoneInfo")) {
						zonemasters.put(Integer.valueOf((String) s.parameters.get(0)), (ZoneMasterConnection) s.parameters.get(1));
						
						ArrayList param = new ArrayList();
						param.add(zones.get(Integer.valueOf((String) s.parameters.get(0))));
						gs.db.addInput(new Signal("GetZoneInfo", param, true));
					}
					if(signal.equals("ZoneInfo")) {
						int zID = (Integer) s.parameters.get(0);
						ArrayList<Zone> adjZones = this.findAdjacentZones(this.zones.get(zID));
						
						zonemasters.get(zID).addInput(s); // Send ZoneInfo to new ZoneMaster
						
						ArrayList params = new ArrayList();
						params.add(this.zones.get(zID));
						params.add(zonemasters.get(zID).getZoneMasterIP());
						
						for(int i = 0; i < adjZones.size(); i++) {
							// Inform existing zonemasters of new adjacent
							ZoneMasterConnection zm = zonemasters.get(adjZones.get(i).ZoneID);
							zm.addInput(new Signal("NewAdjacentZoneMaster", params, true));
							
							// Inform new zonemaster of existing adjacent ones
							ArrayList par = new ArrayList();
							par.add(this.zones.get(adjZones.get(i).ZoneID));
							par.add(this.zonemasters.get(adjZones.get(i).ZoneID).getZoneMasterIP());
							
							zonemasters.get(zID).addInput(new Signal("NewAdjacentZoneMaster", par, true));
						}
						
						ArrayList clientsInZone = getClientsByZone(zID);
						
						for(int i = 0; i < clientsInZone.size(); i++) {
							String name = (String) clientsInZone.get(i);
							ArrayList param = new ArrayList();
							
							param.add(zonemasters.get(zID).getZoneMasterIP());
							param.add("Magic");
							param.add(4510+zID);
							((ClientConnection) clients.get(name)).addInput(new Signal("NewZoneMaster", param, true));
							
							param = new ArrayList();
							param.add("Magic");
							param.add(((ClientConnection) clients.get(name)).getClientIP());
							param.add(name);
		
							zonemasters.get(getZoneMasterByPlayer(currentPCs.get(name))).addInput(new Signal("ChallengeAnswer", param, true));
						}
						
					}
					if(signal.equals("ObjectAdd")) {
						gs.db.addInput(s);
					}
					if(signal.equals("GetObject")) {
						gs.db.addInput(s);
					}
					if(signal.equals("Backup")) {
						String update = (String) s.parameters.get(0);
						String[] updates = update.split(";");
						for(int i = 0; i < updates.length; i++) {
							String[] change = (updates[i]).split(":");
							PlayerCharacter pc = currentPCs.get(change[0]);
							pc.posX = Double.parseDouble(change[1]);
							pc.posY = Double.parseDouble(change[2]);
							pc.posZ = Double.parseDouble(change[3]);
						}
						gs.db.addInput(s);
					}
					if(signal.equals("ClientHasNewMaster")) {
						PlayerCharacter pc = (PlayerCharacter) s.parameters.get(0);
						this.currentPCs.put(pc.name, pc);
						
						ArrayList param = new ArrayList();
						param.add("Magic");
						param.add(clients.get(pc.name).getClientIP());
						param.add(pc.name);
	
						zonemasters.get(getZoneMasterByPlayer(pc)).addInput(new Signal("ChallengeAnswer", param, true));
					}
					if(signal.equals("ZMTimeout")) {
						
					}
					if(signal.equals("ClientTimeout")) {
						
					}
					if(signal.equals("ClientConnected")) {
						ZoneMasterConnection zm = zonemasters.get(getZoneMasterByPlayer(this.currentPCs.get((String) s.parameters.get(0))));
						ArrayList param = new ArrayList();
						param.add(this.currentPCs.get((String) s.parameters.get(0)));
	
						zm.addInput(new Signal("ObjectAdd", param, true));			// Send ObjectAdd(Player's Character) to ZoneMaster
						
						if(this.clients.size() >= 5) {
							for(int i = 0; i < numberOfZones; i++) {
								if(i == 1 && this.clients.size() < 6)
									break;
								if(i == 2 && this.clients.size() < 7)
									break;
								if(i == 3 && this.clients.size() < 8)
									break;
								if(gsZones[i] != null) {
									boolean foundClient = false;
									String player = "";
									int retries = 0;
									while(!foundClient && retries < 10) {
										Set<String> clients = this.clients.keySet();
										Object[] keys =  clients.toArray();
										
										int p = (int)(Math.random() * clients.size());
										player = (String) keys[p];
										if(!masterclients.containsKey(player)) {
											foundClient = true; 
											System.out.println("Found player to manage ZONE " + i + ": " + player);
										}
										retries++;
									}
									if(foundClient) {
										ClientConnection cc = this.clients.get(player);
										param = new ArrayList();
										param.add(zones.get(i));
										cc.addInput(new Signal("ManageZone", param, true));
										masterclients.put(player, i);
	
										this.zonemasters.get(i).addInput(new Signal("StopManagingZone", null, true));
										//gsZones[i].running = false;
										gsZones[i] = null;
									}
								}
							}
						} else {
							
						}
					}
					if(signal.equals("AddObject")) {
						
					}
					if(signal.equals("Disconnect")) { 
						String name = (String) s.parameters.get(0);	// Disconnecting Player's Character Name
						PlayerCharacter pc = currentPCs.get(name);	// Disconnecting Player's Character Object
						this.clients.remove(name);					// Remove connection from list
						this.currentPCs.remove(name);
						ZoneMasterConnection zmc = zonemasters.get(this.getZoneMasterByPlayer(pc));	// Get Player's Zone Master Connection
						ArrayList param = new ArrayList();
						param.add(name);
						zmc.addInput(new Signal("ClientDisconnect", param, true));	// Send Disconnect Signal
						System.out.println("ZoneManager removed player " + pc.name + " from lists..");
						
						if(this.clients.size() < 5) {
							for(int i = 0; i < numberOfZones; i++) {
								gsZones[i] = new GameServerConnection(GameServer.localIP, GameServer.zmchPort, zones.get(0), 4510);
								gsZones[i].start();
							}
							if(masterclients.containsKey(name)) 
								masterclients.remove(name);
						} else {	
							if(masterclients.containsKey(name)) {
								int zoneID = masterclients.get(name);
								masterclients.remove(name);
								if(gsZones[zoneID] == null) {
									boolean foundClient = false;
									String player = "";
									int retries = 0;
									while(!foundClient && retries < 10) {
										Set<String> clients = this.clients.keySet();
										Object[] keys =  clients.toArray();
										
										int p = (int)(Math.random() * clients.size());
										player = (String) keys[p];
										if(!masterclients.containsKey(player)) {
											foundClient = true; 
											System.out.println("Found player to manage ZONE " + zoneID + ": " + player);
										}
										retries++;
									}
									if(foundClient) {
										ClientConnection cc = this.clients.get(player);
										param = new ArrayList();
										param.add(zones.get(zoneID));
										cc.addInput(new Signal("ManageZone", param, true));
										masterclients.put(player, zoneID);
									} else {
										gsZones[zoneID] = new GameServerConnection(GameServer.localIP, GameServer.zmchPort, zones.get(zoneID), 4510 + zoneID);
										gsZones[zoneID].start();
									}
								} else {
									
								}
							}
						}
					}
				}
			}
		}
	}

	/*
	 * Adds a signal to the back of the signal queue.
	 * 
	 * @param input The signal to be added
	 */
	public synchronized void addInput(Signal input) {
		synchronized(this.input) {
			this.input.add(input);
		}
	}

	/*
	 * Returns the Zones that are adjacent to a Zone.
	 * 
	 * @param z		The Zone for which the adjacent zones are requested.
	 * @return		ArrayList containing the Zone objects for adjacent Zones.
	 */
	private ArrayList<Zone> findAdjacentZones(Zone z) {
		ArrayList<Zone> result = new ArrayList<Zone>();
		
		/*** STATIC ZONES ***/
		//int zones = this.zones.size();
		//int zonesInEdge = (int) Math.sqrt(zones);

		if(z.ZoneID == 0) {
			result.add(this.zones.get(1));
			result.add(this.zones.get(2));
			result.add(this.zones.get(3));
			return result;
		}
		if(z.ZoneID == 1) {
			result.add(this.zones.get(0));
			result.add(this.zones.get(2));
			result.add(this.zones.get(3));
			return result;
		}
		if(z.ZoneID == 2) {
			result.add(this.zones.get(0));
			result.add(this.zones.get(1));
			result.add(this.zones.get(3));
			return result;
		}
		if(z.ZoneID == 3) {
			result.add(this.zones.get(0));
			result.add(this.zones.get(1));
			result.add(this.zones.get(2));
			return result;
		}
		
		return result;
	}
	
	/*
	 * Finds the zonemaster of a player.
	 * 
	 * @param pc	PlayerCharacter object of player
	 * @return		ZoneID	for the zone the player is in
	 */
	public int getZoneMasterByPlayer(PlayerCharacter pc) {
		for(int i = 0; i < zones.size(); i++) {
			if(zones.get(i).x <= pc.posX && zones.get(i).y <= pc.posZ) {
				if(zones.get(i).x + zones.get(i).w > pc.posX && zones.get(i).y + zones.get(i).h > pc.posZ) {
					return i;
				}
			}
		}
		System.err.println("Could not find zonemaster for player position!!");
		return -1;
	}
	
	/*
	 * Find connections to clients in a spesific zone
	 * 
	 * @param z	ID of zone
	 * @return ArrayList of ClientConnections
	 */
	public ArrayList getClientsByZone(int z) {
		ArrayList<String> ret = new ArrayList<String>();
		Zone zone = zones.get(z);
		if(zone == null) return null;
		
		Set<String> currentPCs = this.currentPCs.keySet();
		Iterator<String> clIter = currentPCs.iterator();
		while(clIter.hasNext()) {
			String cnm = clIter.next();
			PlayerCharacter pc = this.currentPCs.get(cnm);
			if(pc.posX >= zone.x && pc.posX < zone.x + zone.w) {
				if(pc.posZ >= zone.y && pc.posZ < zone.y + zone.h) {
					ret.add(pc.name);
				}
			}
		}
		
		return ret;
	}
	
	/*
	 * Class showing the zone map
	 */
	public class DrawingPanel extends JPanel {
		ZoneManager zm;
		DrawingPanel (ZoneManager zm) {
		    setBackground (Color.WHITE);
		    this.zm = zm;
		}

		public void paintComponent(Graphics g)   {
		   super.paintComponent(g);
		   g.setColor (Color.BLACK);
		   int dy = getSize ().height;
		   int dx = getSize ().width;
		   
		   for(int i = 0; i < zm.zones.size(); i++) {
			   Zone z = zm.zones.get(i);
			   
			   g.drawRect (z.x*4, z.y*4, z.w*4, z.h*4);
		   }
		}
	}
}
