/**
 * The ZoneMasterHandler class sets up the connection between the ZoneMaster and the 
 * GameClientm, and forwards signals between the two actors over the network.
 */

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.net.Socket;
import java.util.ArrayList;

import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

import com.sun.org.apache.xerces.internal.parsers.DOMParser;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;


public class ZoneMasterHandler extends Thread {

	private ArrayList<SDLTimer> timers = new ArrayList<SDLTimer>(10);
	private ArrayList<Signal> input = new ArrayList<Signal>(100);
	private Socket zonemaster;
	private GameClient gc;
	private String zmip;
	private int zmport;
	private SDLTimer t;
	
	/**
	 * Constructor that sets up a socket connection based on GameClient, IP and port.
	 * @param g
	 * The GameClient, g, that wants to connect to ZoneMaster
	 * @param ip
	 * The ip-address to the ZoneMaster 
	 * @param port
	 * The port to the ZoneMaster
	 */
	public ZoneMasterHandler (GameClient g, String ip, int port){
		this.gc = g;
		this.zmip = ip;
		zmport = port;
		try {
			zonemaster = new Socket(zmip, port);
		} catch (IOException e) {
			System.err.println(e);
		}
	}
	
	/**
	 * Starts the state machine for the ZoneMasterHandler, check SDL design
	 */
	public void run() {
		System.out.println("ZoneMasterHandler starting... Connecting to " + zmip + ":" + zmport);
		try {
			Thread.sleep(2000);
		} catch (InterruptedException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		String state = "Started";
		boolean running = true;
		boolean isEmpty = true;
		try {
			BufferedReader network = new BufferedReader(new InputStreamReader(zonemaster.getInputStream()));

			while(running) {
				while(isEmpty) { // While input queue has no input
					Thread.sleep(10);
					if(network.ready()) {
						String network_in = network.readLine();
						network_in.trim();
						if(!network_in.equals("")) {
							String[] signal_in = network_in.split(";");
							ArrayList netInParam = new ArrayList();
							for(int i = 1; i < signal_in.length; i++) {
								netInParam.add(signal_in[i].trim());
							}
							this.addInput(new Signal(signal_in[0].trim(), netInParam, false));
						}
					}
					for(int i = 0; i < timers.size(); i++) {
						if(timers.get(i).isExpired()) {
							this.addInput(timers.remove(i).getSignal()); //TODO should be added with priority
						}
					}
					synchronized(input) { // protect variable
						if(!input.isEmpty()) isEmpty = false;
					}
				}
				isEmpty = true;
				synchronized(input) {
					Signal s = input.remove(0);
					String signal = s.name;
					System.out.println("ZoneMasterHandler received signal: " + signal + " in state " + state);
					if(state.equals("Started")) {
						if(signal.equals("ConnectZM")) {
							signalGS(s);
						String a =	(String)s.parameters.get(0);
						String b =	(String)s.parameters.get(1);
						String c =	(String)s.parameters.get(2);
						System.out.println("Param 1: " + a + " Param 2: " + b + " Param 3: " + c);
							
							state = "ZMConnecting";
							}
						}
					else if(state.equals("ZMConnecting")){
						if(signal.equals("ConnectZMOK")){
							gc.addInput(s);
							t = new SDLTimer(60000, new Signal("KeepAlive", null, false));
							t.start();
							timers.add(t);
							state = "Connected";
						}
						if(signal.equals("ConnectZMFail")){
							gc.addInput(s);
							state = "Started";
						}
					}
					else if(state.equals("Connected")){
						
						if(signal.equals("NewZoneMaster")){
							gc.addInput(s);
							running = false;
						}
						if(signal.equals("ChangeState")){
							if(s.internal){
								signalGS(s);
								state = "Connected";		
							}
							else{
								
								s.internal = true;
								gc.addInput(s);
								state = "Connected";
						}
					}
						if(signal.equals("CallProcedure")){
							if(s.internal){
								signalGS(s);
								state = "Connected";
							}
							else{
								s.internal = true;
								gc.addInput(s);
								state = "Connected";
						}
					}
						if(signal.equals("ObjectAdd")){
							if(s.internal){
								System.err.println("FEIL");
							}
							else{
								System.out.println(s.parameters.get(0).toString());
								
								GameObjectList go = parseXMLString((String) s.parameters.get(0) );
								ArrayList param = new ArrayList();
								param.add(go);
								gc.addInput(new Signal("ObjectAdd", param, true));

								
								/*Dummy signal for sjekk av change state
								param = new ArrayList();
								param.add(0);
								param.add("moving");
								param.add("true");
								
								signalGS(new Signal("ChangeState",param, false));
							
								
								param = new ArrayList();
								param.add(0);
								param.add("pickUp");
								param.add(1321);
								signalGS(new Signal("CallProcedure",param, false));
								*/
								state = "Connected";
							}
						}
						if(signal.equals("AOIUpdate")){
								ArrayList param = new ArrayList();
								param.add(parseXMLString((String)s.parameters.get(0)));
								gc.addInput(new Signal("AOIUpdate", param, false));
								
								state = "Connected";	
							
						}
						if(signal.equals("KeepAlive")){
							t.reset();
						}
					}
				}
			}
		}
		
				catch (IOException e) {
					System.err.println(e);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
					
									
					/**
					 * Adds a signal from the external GameServer to an internal queue.
					 * 
					 * @param input
					 * The signal that will be added to the input queue.
					 */	
					public void addInput(Signal input) {
						synchronized(this.input) {
							this.input.add(input);
						}
					}
					
					/**
					 * Sends a signal over the network on the OutputStream.
					 * 
					 * @param output
					 * The signal that will be sent over the network.
					 */
					public void signalGS(Signal output) {
						try {
							PrintStream os = new PrintStream(zonemaster.getOutputStream());
							String out = output.name;
							if(output.parameters != null)
								for(int i = 0; i < output.parameters.size(); i++) {
									out = out + ";" + output.parameters.get(i).toString();
								}
							os.print(out + "\n");
						} catch (IOException e) {
							System.err.println(e);
						}
					}
					
					/**
					 * Deserialises the XML string to a PlayerCharacter object.
					 * 
					 * @param xml
					 * The xml string that is sent over the network
					 * 
					 * @return
					 * The PlayerCharacter object
					 */
					public GameObjectList parseXMLString(String xml) {
						GameObjectList gol = new GameObjectList();
						try {
							DOMParser parser = new DOMParser();
							parser.parse(new InputSource(new java.io.StringReader(xml)));
							Document doc = parser.getDocument();

							doc.getDocumentElement().normalize();
							
							NodeList nodeLst = doc.getElementsByTagName("GameObject");
							for (int s = 0; s < nodeLst.getLength(); s++) {
								Node fstNode = nodeLst.item(s);
								if (fstNode.getNodeType() == Node.ELEMENT_NODE) {			  
									Element fstElmnt = (Element) fstNode;
									
									NodeList oidElmntLst = fstElmnt.getElementsByTagName("objectID");
									Element oidElmnt = (Element) oidElmntLst.item(0);
									NodeList oid = oidElmnt.getChildNodes();
									int oid_ = Integer.parseInt(((Node) oid.item(0)).getNodeValue());
									
									NodeList typeElmntLst = fstElmnt.getElementsByTagName("type");
									Element typeElmnt = (Element) typeElmntLst.item(0);
									NodeList type = typeElmnt.getChildNodes();
									String type_ = ((Node) type.item(0)).getNodeValue();
									
									NodeList posXElmntLst = fstElmnt.getElementsByTagName("posX");
									Element posXElmnt = (Element) posXElmntLst.item(0);
									NodeList posX = posXElmnt.getChildNodes();
									double posX_ = Double.parseDouble(((Node) posX.item(0)).getNodeValue());
									
									NodeList posYElmntLst = fstElmnt.getElementsByTagName("posY");
									Element posYElmnt = (Element) posYElmntLst.item(0);
									NodeList posY = posYElmnt.getChildNodes();
									double posY_ = Double.parseDouble(((Node) posY.item(0)).getNodeValue());
									
									NodeList posZElmntLst = fstElmnt.getElementsByTagName("posZ");
									Element posZElmnt = (Element) posZElmntLst.item(0);
									NodeList posZ = posZElmnt.getChildNodes();
									double posZ_ = Double.parseDouble(((Node) posZ.item(0)).getNodeValue());
									
									NodeList rotXElmntLst = fstElmnt.getElementsByTagName("rotX");
									Element rotXElmnt = (Element) rotXElmntLst.item(0);
									NodeList rotX = rotXElmnt.getChildNodes();
									double rotX_ = Double.parseDouble(((Node) rotX.item(0)).getNodeValue());
									
									NodeList rotYElmntLst = fstElmnt.getElementsByTagName("rotY");
									Element rotYElmnt = (Element) rotYElmntLst.item(0);
									NodeList rotY = rotYElmnt.getChildNodes();
									double rotY_ = Double.parseDouble(((Node) rotY.item(0)).getNodeValue());
									
									NodeList rotZElmntLst = fstElmnt.getElementsByTagName("rotZ");
									Element rotZElmnt = (Element) rotZElmntLst.item(0);
									NodeList rotZ = rotZElmnt.getChildNodes();
									double rotZ_ = Double.parseDouble(((Node) rotZ.item(0)).getNodeValue());
									
									GameObject go = new GameObject(oid_, type_, posX_, posY_, posZ_, rotX_, rotY_, rotZ_);
									gol.add(go);
								}
							}
							
							nodeLst = doc.getElementsByTagName("PlayerCharacter");
							for (int s = 0; s < nodeLst.getLength(); s++) {
								Node fstNode = nodeLst.item(s);
								if (fstNode.getNodeType() == Node.ELEMENT_NODE) {			  
									Element fstElmnt = (Element) fstNode;
									
									NodeList oidElmntLst = fstElmnt.getElementsByTagName("objectID");
									Element oidElmnt = (Element) oidElmntLst.item(0);
									NodeList oid = oidElmnt.getChildNodes();
									int oid_ = Integer.parseInt(((Node) oid.item(0)).getNodeValue());
									
									NodeList posXElmntLst = fstElmnt.getElementsByTagName("posX");
									Element posXElmnt = (Element) posXElmntLst.item(0);
									NodeList posX = posXElmnt.getChildNodes();
									double posX_ = Double.parseDouble(((Node) posX.item(0)).getNodeValue());
									
									NodeList posYElmntLst = fstElmnt.getElementsByTagName("posY");
									Element posYElmnt = (Element) posYElmntLst.item(0);
									NodeList posY = posYElmnt.getChildNodes();
									double posY_ = Double.parseDouble(((Node) posY.item(0)).getNodeValue());
									
									NodeList posZElmntLst = fstElmnt.getElementsByTagName("posZ");
									Element posZElmnt = (Element) posZElmntLst.item(0);
									NodeList posZ = posZElmnt.getChildNodes();
									double posZ_ = Double.parseDouble(((Node) posZ.item(0)).getNodeValue());
									
									NodeList rotXElmntLst = fstElmnt.getElementsByTagName("rotX");
									Element rotXElmnt = (Element) rotXElmntLst.item(0);
									NodeList rotX = rotXElmnt.getChildNodes();
									double rotX_ = Double.parseDouble(((Node) rotX.item(0)).getNodeValue());
									
									NodeList rotYElmntLst = fstElmnt.getElementsByTagName("rotY");
									Element rotYElmnt = (Element) rotYElmntLst.item(0);
									NodeList rotY = rotYElmnt.getChildNodes();
									double rotY_ = Double.parseDouble(((Node) rotY.item(0)).getNodeValue());
									
									NodeList rotZElmntLst = fstElmnt.getElementsByTagName("rotZ");
									Element rotZElmnt = (Element) rotZElmntLst.item(0);
									NodeList rotZ = rotZElmnt.getChildNodes();
									double rotZ_ = Double.parseDouble(((Node) rotZ.item(0)).getNodeValue());
									
									NodeList nameElmntLst = fstElmnt.getElementsByTagName("name");
									Element nameElmnt = (Element) nameElmntLst.item(0);
									NodeList name = nameElmnt.getChildNodes();
									String name_ = ((Node) name.item(0)).getNodeValue();
									
									NodeList movingElmntLst = fstElmnt.getElementsByTagName("moving");
									Element movingElmnt = (Element) movingElmntLst.item(0) ;
									NodeList moving = movingElmnt.getChildNodes();
									boolean moving_ = Boolean.parseBoolean(((Node) moving.item(0) ).getNodeValue());
									
									PlayerCharacter go = new PlayerCharacter(oid_, posX_, posY_, posZ_, rotX_, rotY_, rotZ_, name_);
									go.moving = moving_;
									gol.add(go);
								}
							}
							
						} catch (Exception e) {
							e.printStackTrace();
						}
						return gol;
					}
				}
		
				
			
