/*
 * Class for managing a zone.
 * Contains references about proxies, players in zone and adjacent zone masters.
 * Handles moving characters and synchronizing actions between players.
 * 
 * @author Frode Voll Aasen
 * @author Tom Christian Bjrlo Johannesen
 * 
 */
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Set;
import java.awt.Color;
import java.awt.Graphics;
import java.lang.reflect.*;

import javax.swing.JFrame;
import javax.swing.JPanel;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;

import com.sun.org.apache.bcel.internal.generic.NEW;
import com.sun.org.apache.xerces.internal.parsers.DOMParser;
import java.io.*;

public class ZoneModel extends Thread {
	private ArrayList<SDLTimer> timers = new ArrayList<SDLTimer>(10);
	GameObjectList gol;
	HashMap<Integer, GameObject> objects = new HashMap<Integer, GameObject>();
	HashMap<String, ZMClientConnection> clients = new HashMap<String, ZMClientConnection>();
	HashMap<String, PlayerCharacter> characters = new HashMap<String, PlayerCharacter>();
	HashMap<Integer, Zone> adjacentZones = new HashMap<Integer, Zone>();
	HashMap<Integer, ZMZMConnection> adjacentZoneMasters = new HashMap<Integer, ZMZMConnection>();
	HashMap<Integer, ArrayList> proxies = new HashMap<Integer, ArrayList>(); // zoneId -> pcIDs
	HashMap<Integer, GameObject> proxiesIn = new HashMap<Integer, GameObject>(); // ObjectID -> object
	
	ZMZMConnectionHandler zmzmc = new ZMZMConnectionHandler(this);
	GameServerConnection gsc;
	
	Zone Zone;
	ThreeDUnit td;
	public boolean running = true;
	
	/*
	 * Class constructor
	 * 
	 * @param Zone	Zone object of the zone
	 * @param gol	GameObjectList containing all the objects in this zone
	 * @param gsc	Reference to the GameServer connection
	 */
	public ZoneModel(Zone Zone, GameObjectList gol, GameServerConnection gsc) {
		this.Zone = Zone;
		this.gol = gol;
		this.gsc = gsc;
		for(int i = 0; i < gol.GameObjects.size(); i++) {
			objects.put(gol.GameObjects.get(i).objectID, gol.GameObjects.get(i));
		}
		this.zmzmc.start();
	}
	
	private ArrayList<Signal> input = new ArrayList<Signal>(100);

	public void run() {
		
		JFrame window = new JFrame("ZoneModel <" + Zone.ZoneID + ">");
		window.setVisible(true);
		DrawingPanel panel = this.new DrawingPanel(this);
		window.setSize(Zone.w*4+7, Zone.h*4+33);
		window.add(panel);
		
		window.setResizable(false);
		
		
		System.out.println("ZoneModel starting...ID: " + Zone.ZoneID);
		
		td = new ThreeDUnit(this, gol);
		td.start();
		
		SDLTimer t = new SDLTimer(10000, new Signal("Backup", null, true));
		t.start();
		timers.add(t);
		
		t = new SDLTimer(2000, new Signal("CheckStates", null, true));
		t.start();
		timers.add(t);
		
		String state = "Started";
		boolean isEmpty = true;
		while(running) {
			while(isEmpty) { // While input queue has no input
				try {
					Thread.sleep(1);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				for(int i = 0; i < timers.size(); i++) {
					if(timers.get(i).isExpired()) {
						this.addInput(timers.remove(i).getSignal()); //TODO should be added with priority
					}
				}
				synchronized(input) { // protect variable
					if(!input.isEmpty()) isEmpty = false;
				}
			}
			isEmpty = true;
			synchronized(input) {
				Signal s = input.remove(0);
				String signal = s.name;
				Date now = new Date();
				if(!signal.equals("CheckStates"))
					System.out.println(now.getTime() + ": ZoneModel (" + Zone.ZoneID + ") received signal: " + signal + " in state " + state);
				if(state.equals("Started")) {
					if(signal.equals("ClientConnected")) {
						System.out.println(now.getTime() + ": ZoneModel (" + Zone.ZoneID + ") Client Connected!");
						this.clients.put((String) s.parameters.get(0), (ZMClientConnection) s.parameters.get(1));
						ArrayList param = new ArrayList();
						param.add((String) s.parameters.get(0));
						t = new SDLTimer(20000, new Signal("AOIUpdate", param, true));
						t.start();
						timers.add(t);
						if(characters.get((String) s.parameters.get(0)) != null) {						
							GameObjectList affectedPlayers = this.findPlayersInAOI((GameObject) characters.get((String) s.parameters.get(0)));
							for(int i = 0; i < affectedPlayers.GameObjects.size(); i++) {
								param = new ArrayList();
								param.add(affectedPlayers.GameObjects.get(i));
								((ZMClientConnection) s.parameters.get(1)).addInput(new Signal("ObjectAdd", param, true));
							}
							
							if(this.proxiesIn.get(characters.get((String) s.parameters.get(0)).objectID) != null) {
								this.proxiesIn.remove(characters.get((String) s.parameters.get(0)).objectID);
							}
						}
					}
					if(signal.equals("ClientTimeout")) {
						PlayerCharacter pc = this.characters.get((String) s.parameters.get(0));
						this.characters.remove((String) s.parameters.get(0));
						this.clients.remove(pc.name);
						this.objects.remove(pc.objectID);
					}
					if(signal.equals("ClientDisconnect")) {
						PlayerCharacter pc = this.characters.get((String) s.parameters.get(0));
						this.characters.remove((String) s.parameters.get(0));
						this.clients.remove(pc.name);
						this.objects.remove(pc.objectID);
						System.out.println("ZoneMaster removed player " + pc.name + " from lists..");
					}
					if(signal.equals("NewAdjacentZoneMaster")) {
						Zone z = (Zone) s.parameters.get(0);
						//this.adjacentZoneMasters.put(z.ZoneID, new ZMZMConnection((String) s.parameters.get(1)));
						this.adjacentZones.put(z.ZoneID, z);
						ZMZMConnection adjZM = new ZMZMConnection(this, (String) s.parameters.get(1), 4500+z.ZoneID);
						adjZM.start();
						if(adjacentZoneMasters.get(z.ZoneID) != null)
							adjacentZoneMasters.get(z.ZoneID).running = false;
						adjacentZoneMasters.put(z.ZoneID, adjZM);
						proxies.put(z.ZoneID, new ArrayList());
					}
					if(signal.equals("CallProcedure")) {
						PlayerCharacter o = (PlayerCharacter) objects.get(Integer.parseInt((String) s.parameters.get(0)));
						try {
							/*Field field = o.getClass().getField((String) s.parameters.get(1));
							String type = field.getType().toString();
							if(type.equals("boolean")) {
								field.setBoolean(o, Boolean.valueOf( (String) s.parameters.get(2) ));
							}*/
							//Class[] parametertype = new Class[] {String.class};
							//Method method = o.getClass().getMethod( (String) s.parameters.get(1), parametertype);
							Method[] methods = o.getClass().getMethods();
							Method method = null;
							Object[] arguments = null;
							
							for(int i = 0; i < methods.length; i++) {
								if(methods[i].getName().equals((String) s.parameters.get(1))) {
									method = methods[i];
									break;
								}
							}
							Class[] parameters = method.getParameterTypes();
							if(parameters[0].getName().equals("int")) {
								arguments = new Object[] { Integer.parseInt((String) s.parameters.get(2))};
							}
							method.invoke(o, arguments);
						} catch (SecurityException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						} catch (IllegalArgumentException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						} catch (IllegalAccessException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						} catch (InvocationTargetException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}
						
						GameObjectList affectedPlayers = this.findPlayersInAOI(o);
						s.internal = true;
						for(int i = 0; i < affectedPlayers.GameObjects.size(); i++) {
							clients.get( ((PlayerCharacter) affectedPlayers.GameObjects.get(i)).name ).addInput(s);
						}
					}
					if(signal.equals("ChangeState")) {
						PlayerCharacter o = (PlayerCharacter) objects.get(Integer.parseInt((String) s.parameters.get(0)));
						if(o != null) {
							try {
								Field field = o.getClass().getField((String) s.parameters.get(1));
								String type = field.getType().toString();
								if(type.equals("boolean")) {
									field.setBoolean(o, Boolean.valueOf( (String) s.parameters.get(2) ));
								}
								if(type.equals("double")) {
									field.setDouble(o, Double.valueOf( (String) s.parameters.get(2) ));
								}
							} catch (SecurityException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							} catch (NoSuchFieldException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							} catch (IllegalArgumentException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							} catch (IllegalAccessException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							}
							if(o.moving) {
								System.out.println(o.name + " is now moving");
							} else {
								System.out.println(o.name + " is now standing still");
							}
							
							GameObjectList affectedPlayers = this.findPlayersInAOI(o);
							s.internal = true;
							for(int i = 0; i < affectedPlayers.GameObjects.size(); i++) {
								if(!o.name.equals(((PlayerCharacter) affectedPlayers.GameObjects.get(i)).name))
									clients.get( ((PlayerCharacter) affectedPlayers.GameObjects.get(i)).name ).addInput(s);
							}
							
							Set<Integer> adjZKeys = this.proxies.keySet();
							Iterator<Integer> adjZIter = adjZKeys.iterator();
							while(adjZIter.hasNext()) {
								int ZoneID = adjZIter.next();
								ArrayList px = proxies.get(ZoneID);
								
								if(px.contains((Object) o.objectID)) {
									System.out.println("ZoneModel (" + Zone.ZoneID + ") Sending update proxy to zone " + ZoneID);
									ZMZMConnection proxymaster = adjacentZoneMasters.get(ZoneID);
									ArrayList param = new ArrayList();
									param.add(o.objectID);
									param.add("ChangeState");
									param.add(s.parameters.get(1));
									param.add(s.parameters.get(2));
									proxymaster.addInput(new Signal("UpdateProxy", param, true));
								}
							}
							
						}
					}
					if(signal.equals("AOIUpdate")) {
						ZMClientConnection z = clients.get(s.parameters.get(0));
						PlayerCharacter p = characters.get(s.parameters.get(0));
						
						if(p != null) {
							GameObjectList affectedPlayers = this.findPlayersInAOI(p);
							ArrayList param = new ArrayList();
							param.add(affectedPlayers);
							z.addInput(new Signal("AOIUpdate", param, false));
							
							param = new ArrayList();
							param.add(p.name);
							t = new SDLTimer(20000, new Signal("AOIUpdate", param, true));
							t.start();
							timers.add(t);
						}
					}
					if(signal.equals("ObjectAdd")) {
						if(((GameObject) ((GameObjectList) s.parameters.get(0)).GameObjects.get(0)).type.equals("PlayerCharacter")) {
							String name = ((PlayerCharacter) ((GameObjectList) s.parameters.get(0)).GameObjects.get(0)).name;
							PlayerCharacter pc = (PlayerCharacter) ((GameObjectList) s.parameters.get(0)).GameObjects.get(0);
							characters.put(name, pc);
							ZMClientConnection c = clients.get(name);
						}
						objects.put(((GameObject) ((GameObjectList) s.parameters.get(0)).GameObjects.get(0)).objectID, (GameObject) ((GameObjectList) s.parameters.get(0)).GameObjects.get(0));
						
						GameObjectList affectedPlayers = this.findPlayersInAOI((GameObject) ((GameObjectList) s.parameters.get(0)).GameObjects.get(0));
						s.internal = true;
						for(int i = 0; i < affectedPlayers.GameObjects.size(); i++) {
							ZMClientConnection c = clients.get( ((PlayerCharacter) affectedPlayers.GameObjects.get(i)).name );
							if(c != null) c.addInput(s);
						}
					}
					if(signal.equals("CreateProxy")) {
						PlayerCharacter pc = (PlayerCharacter) parseXMLString((String) s.parameters.get(0)).GameObjects.get(0);
						System.out.println("Creating PROXY of object "+ pc.name +"!");
						proxiesIn.put(pc.objectID, pc);
						//characters.put(pc.name, pc);
					}
					if(signal.equals("UpdateProxy")) {
						int oid = Integer.parseInt((String)s.parameters.get(0));
						String update = (String) s.parameters.get(1);
						String p1 = (String) s.parameters.get(2);
						String p2 = (String) s.parameters.get(3);
						
						if(update.equals("ChangeState")) {
							PlayerCharacter o = (PlayerCharacter) proxiesIn.get(oid);
							if(o != null) {
								try {
									Field field = o.getClass().getField(p1);
									String type = field.getType().toString();
									if(type.equals("boolean")) {
										field.setBoolean(o, Boolean.valueOf(p2));
									}
									if(type.equals("double")) {
										field.setDouble(o, Double.valueOf(p2));
									}
								} catch (SecurityException e) {
									// TODO Auto-generated catch block
									e.printStackTrace();
								} catch (NoSuchFieldException e) {
									// TODO Auto-generated catch block
									e.printStackTrace();
								} catch (IllegalArgumentException e) {
									// TODO Auto-generated catch block
									e.printStackTrace();
								} catch (IllegalAccessException e) {
									// TODO Auto-generated catch block
									e.printStackTrace();
								}
								if(o.moving) {
									System.out.println(o.name + "(proxy) is now moving");
								} else {
									System.out.println(o.name + "(proxy) is now standing still");
								}
								
								GameObjectList affectedPlayers = this.findPlayersInAOI(o);
								ArrayList param = new ArrayList();
								param.add(oid);
								param.add(p1);
								param.add(p2);
								for(int i = 0; i < affectedPlayers.GameObjects.size(); i++) {
									if(!o.name.equals(((PlayerCharacter) affectedPlayers.GameObjects.get(i)).name))
										clients.get( ((PlayerCharacter) affectedPlayers.GameObjects.get(i)).name ).addInput(new Signal("ChangeState", param, true));
								}
							}
						}
						else if(update.equals("CallProcedure")) {
							
						}
					}
					if(signal.equals("Backup")) {
						Set<String> chkeys = this.characters.keySet();
						
						Iterator<String> i = chkeys.iterator();
						String update = "";
						while(i.hasNext()) {
							String key = i.next();
							PlayerCharacter pc = this.characters.get(key);
							update += pc.name + ":" + pc.posX + ":" + pc.posY + ":" + pc.posZ + ";";
						}
						if(update != "") {
							ArrayList param = new ArrayList();
							param.add(update);
							this.gsc.addInput(new Signal("Backup", param, true));
						}
						t = new SDLTimer(10000, new Signal("Backup", null, true));
						t.start();
						timers.add(t);
					}
					if(signal.equals("CheckStates")) {
						Set<String> keys = this.characters.keySet();
						
						checkProxies();

						Iterator<String> i = keys.iterator();
						while(i.hasNext()) {
							String key = i.next();
							PlayerCharacter pc = this.characters.get(key);
							if(clients.get(pc.name) == null) {
								System.err.println("No connection to client, aborting..");
								break;
							}
							now = new Date();
							long currentTime = now.getTime();
							
							if(pc.moving) {
								if(pc.lastStateCheck > 0) {
									long deltaTime = currentTime - pc.lastStateCheck;
									
									pc.posX += ((double)deltaTime/1000)*(1)*Math.sin(pc.rotY);
									pc.posZ += ((double)deltaTime/1000)*(-1)*Math.cos(pc.rotY);
									
									boolean inZone = true;
									
									if(pc.posX >= Zone.x + Zone.w) { // Player leaving zone at right edge.
										if(Zone.x + Zone.w == GameServer.worldWidth) pc.posX -= ((double)deltaTime/1000)*(3)*Math.sin(pc.rotY);
										else {
											ZMZMConnection newZM = null;
											Zone z = null;
											Set<Integer> adjZKeys = this.adjacentZones.keySet();
											Iterator<Integer> adjZIter = adjZKeys.iterator();
											while(adjZIter.hasNext()) {
												int ZoneID = adjZIter.next();
												z = adjacentZones.get(ZoneID);
												if(z.x >= Zone.x+Zone.w && z.y <= pc.posZ && z.y + z.h > pc.posZ) {
													newZM = adjacentZoneMasters.get(z.ZoneID);
													break;
												}
											}
											if(newZM != null && z != null) {
												ArrayList param = new ArrayList();
												param.add(pc);
												param.add(z);
												this.addInput(new Signal("ClientLeftZone", param, true));
												this.proxies.get(z.ZoneID).remove((Object)pc.objectID);
											} else {
												System.err.println("Error: Client Left Zone But No ZoneMaster Found!!!!");
											}
										}
										inZone = false;
									} else if(pc.posX >= Zone.x + Zone.w - 15 && Zone.x + Zone.w != GameServer.worldWidth) { // Player entering buffer at right edge
										ZMZMConnection newZM = null;
										Zone z = null;
										Set<Integer> adjZKeys = this.adjacentZones.keySet();
										Iterator<Integer> adjZIter = adjZKeys.iterator();
										while(adjZIter.hasNext()) {
											int ZoneID = adjZIter.next();
											z = adjacentZones.get(ZoneID);
											if(z.x >= Zone.x+Zone.w && z.y <= pc.posZ && z.y + z.h > pc.posZ) {
												newZM = adjacentZoneMasters.get(z.ZoneID);
												break;
											}
										}
										if(newZM != null && z != null) {
											if(!proxies.get(z.ZoneID).contains(pc.objectID)) {
												ArrayList param = new ArrayList();
												param.add(pc);
												newZM.addInput(new Signal("CreateProxy", param, true));
												proxies.get(z.ZoneID).add(pc.objectID);
											}
										} else {
											System.err.println("Error: Client entered buffer but no zonemaster to create proxy!!!!");
										}
										inZone = false;
									}
									if(pc.posX < Zone.x) { // Player leaving zone at left edge
										if(Zone.x == 0) pc.posX -= ((double)deltaTime/1000)*(3)*Math.sin(pc.rotY);
										else {
											ZMZMConnection newZM = null;
											Zone z = null;
											Set<Integer> adjZKeys = this.adjacentZones.keySet();
											Iterator<Integer> adjZIter = adjZKeys.iterator();
											while(adjZIter.hasNext()) {
												int ZoneID = adjZIter.next();
												z = adjacentZones.get(ZoneID);
												if(z.x + z.w <= Zone.x && z.y <= pc.posZ && z.y + z.h > pc.posZ) {
													newZM = adjacentZoneMasters.get(z.ZoneID);
													break;
												}
											}
											if(newZM != null && z != null) {
												ArrayList param = new ArrayList();
												param.add(pc);
												param.add(z);
												this.addInput(new Signal("ClientLeftZone", param, true));
											} else {
												System.err.println("Error: Client Left Zone But No ZoneMaster Found!!!!");
											}
										}
										inZone = false;
									} else if(pc.posX < Zone.x + 15 && Zone.x != 0) { // Player entering buffer at left edge
										ZMZMConnection newZM = null;
										Zone z = null;
										Set<Integer> adjZKeys = this.adjacentZones.keySet();
										Iterator<Integer> adjZIter = adjZKeys.iterator();
										while(adjZIter.hasNext()) {
											int ZoneID = adjZIter.next();
											z = adjacentZones.get(ZoneID);
											if(z.x + z.w <= Zone.x && z.y <= pc.posZ && z.y + z.h > pc.posZ) {
												newZM = adjacentZoneMasters.get(z.ZoneID);
												break;
											}
										}
										if(newZM != null && z != null) {
											if(!proxies.get(z.ZoneID).contains(pc.objectID)) {
												ArrayList param = new ArrayList();
												param.add(pc);
												newZM.addInput(new Signal("CreateProxy", param, true));
												proxies.get(z.ZoneID).add(pc.objectID);
											}
										} else {
											System.err.println("Error: Client entered buffer but no zonemaster to create proxy!!!!");
										}
										inZone = false;
									} 
									if(pc.posZ >= Zone.y + Zone.h) { // Player leaving zone at bottom edge
										if(Zone.y + Zone.h == GameServer.worldHeight) pc.posZ -= ((double)deltaTime/1000)*(-5)*Math.cos(pc.rotY);
										else {
											ZMZMConnection newZM = null;
											Zone z = null;
											Set<Integer> adjZKeys = this.adjacentZones.keySet();
											Iterator<Integer> adjZIter = adjZKeys.iterator();
											while(adjZIter.hasNext()) {
												int ZoneID = adjZIter.next();
												z = adjacentZones.get(ZoneID);
												if(z.y >= Zone.y + Zone.h && z.x <= pc.posX && z.x + z.w > pc.posX) {
													newZM = adjacentZoneMasters.get(z.ZoneID);
													break;
												}
											}
											if(newZM != null && z != null) {
												ArrayList param = new ArrayList();
												param.add(pc);
												param.add(z);
												this.addInput(new Signal("ClientLeftZone", param, true));
											} else {
												System.err.println("Error: Client Left Zone But No ZoneMaster Found!!!!");
											}
										}
										inZone = false;
									} else if(pc.posZ > Zone.y + Zone.h - 15 && Zone.y + Zone.h != GameServer.worldHeight) { // Player entering buffer at bottom edge
										ZMZMConnection newZM = null;
										Zone z = null;
										Set<Integer> adjZKeys = this.adjacentZones.keySet();
										Iterator<Integer> adjZIter = adjZKeys.iterator();
										while(adjZIter.hasNext()) {
											int ZoneID = adjZIter.next();
											z = adjacentZones.get(ZoneID);
											if(z.y >= Zone.y + Zone.h && z.x <= pc.posX && z.x + z.w > pc.posX) {
												newZM = adjacentZoneMasters.get(z.ZoneID);
												break;
											}
										}
										if(newZM != null && z != null) {
											if(!proxies.get(z.ZoneID).contains(pc.objectID)) {
												ArrayList param = new ArrayList();
												param.add(pc);
												newZM.addInput(new Signal("CreateProxy", param, true));
												proxies.get(z.ZoneID).add(pc.objectID);
											}
										} else {
											System.err.println("Error: Client entered buffer but no zonemaster to create proxy!!!!");
										}
										inZone = false;
									}
									if(pc.posZ < Zone.y) { // Player leaving zone at top edge
										if(Zone.y == 0) pc.posZ -= ((double)deltaTime/1000)*(-5)*Math.cos(pc.rotY);
										else {
											ZMZMConnection newZM = null;
											Zone z = null;
											Set<Integer> adjZKeys = this.adjacentZones.keySet();
											Iterator<Integer> adjZIter = adjZKeys.iterator();
											while(adjZIter.hasNext()) {
												int ZoneID = adjZIter.next();
												z = adjacentZones.get(ZoneID);
												if(z.y + z.h <= Zone.y && z.x <= pc.posX && z.x + z.w > pc.posX) {
													newZM = adjacentZoneMasters.get(z.ZoneID);
													break;
												}
											}
											if(newZM != null && z != null) {
												ArrayList param = new ArrayList();
												param.add(pc);
												param.add(z);
												this.addInput(new Signal("ClientLeftZone", param, true));
											} else {
												System.err.println("Error: Client Left Zone But No ZoneMaster Found!!!!");
											}
										}
										inZone = false;
									} else if(pc.posZ < Zone.y + 15 && Zone.y != 0) { // Player entering buffer at top edge
										ZMZMConnection newZM = null;
										Zone z = null;
										Set<Integer> adjZKeys = this.adjacentZones.keySet();
										Iterator<Integer> adjZIter = adjZKeys.iterator();
										while(adjZIter.hasNext()) {
											int ZoneID = adjZIter.next();
											z = adjacentZones.get(ZoneID);
											if(z.y + z.h <= Zone.y && z.x <= pc.posX && z.x + z.w > pc.posX) {
												newZM = adjacentZoneMasters.get(z.ZoneID);
												break;
											}
										}
										if(newZM != null && z != null) {
											if(!proxies.get(z.ZoneID).contains(pc.objectID)) {
												ArrayList param = new ArrayList();
												param.add(pc);
												newZM.addInput(new Signal("CreateProxy", param, true));
												proxies.get(z.ZoneID).add(pc.objectID);
											}
										} else {
											System.err.println("Error: Client entered buffer but no zonemaster to create proxy!!!!");
										}
										inZone = false;
									}
									if(inZone) {	// Player is in zone and not in buffer
										Set<Integer> pKeys = this.proxies.keySet();
										Iterator<Integer> pIter = pKeys.iterator();
										while(pIter.hasNext()) {
											int ZoneID = pIter.next();
											ArrayList prox = (ArrayList) proxies.get(ZoneID);
											if(prox.contains(pc.objectID))
												prox.remove((Object)pc.objectID);
										}
									}
								}
								
								/*
								 * 
								 * PLAYERS
								 * 
								 */
								
								GameObjectList affectedPlayers = this.findPlayersInAOI(pc);
								GameObjectList affectedProxies = this.findProxiesInAOI(pc);
								
								ArrayList<Integer> newPlayers = new ArrayList<Integer>();
								ArrayList<Integer> newProxies = new ArrayList<Integer>();
																
								//Add all players in AOI to newPlayers list
								for(int j = 0; j < affectedPlayers.GameObjects.size(); j++) {
									newPlayers.add(affectedPlayers.GameObjects.get(j).objectID);
									//System.out.println("PC" + pc.objectID + "  In AOI: " + affectedPlayers.GameObjects.get(j).objectID);
								}
								
								//Add all proxies in AOI to newProxies list
								for(int j = 0; j < affectedProxies.GameObjects.size(); j++) {
									newProxies.add(affectedProxies.GameObjects.get(j).objectID);
									//System.out.println("PC" + pc.objectID + "  In AOI: " + affectedPlayers.GameObjects.get(j).objectID);
								}
								
								ArrayList<Integer> removeItems = new ArrayList<Integer>();
								
								// Remove players and proxies that the player thinks is in his AOI, but are no longer there
								for(int j = 0; j < pc.AOIObjects.size(); j++) {
									if(!newPlayers.contains(pc.AOIObjects.get(j)) && !newProxies.contains(pc.AOIObjects.get(j))) {
										//System.out.println("PC" + pc.objectID + "  Obj " + pc.AOIObjects.get(j) + " removed (inexistent)");
										removeItems.add(j);
									}
								}
								int skew = 0;
								for(int j = 0; j < removeItems.size(); j++) {
									pc.AOIObjects.remove((int)removeItems.get(j)-skew);
									skew++;
								}
								
								removeItems = new ArrayList<Integer>();
								
								// If player already knows the player is in his AOI, remove player.
								// If not add player to players AOIObjects list.
								for(int j = 0; j < newPlayers.size(); j++) {
									if(pc.AOIObjects.contains(newPlayers.get(j))) {
										//System.out.println("PC" + pc.objectID + "  Obj " + newPlayers.get(j) + " removed (already exists)");
										removeItems.add(j);
									} else {
										pc.AOIObjects.add(newPlayers.get(j));
										//System.out.println("PC" + pc.objectID + " Obj " + newPlayers.get(j) + " added");
									}
								}
								skew = 0;
								for(int j = 0; j < removeItems.size(); j++) {
									newPlayers.remove((int) removeItems.get(j)-skew);
									skew++;
								}
								
								s.internal = true;
								for(int i1 = 0; i1 < newPlayers.size(); i1++) {
									ZMClientConnection c = clients.get( ((PlayerCharacter) (objects.get(newPlayers.get(i1)))).name );
									
									// Notify other player of moving player entering AOI
									if(!((PlayerCharacter) objects.get(newPlayers.get(i1))).name.equals(pc.name)) {
										ArrayList param = new ArrayList();
										param.add((PlayerCharacter) pc);
										c.addInput(new Signal("ObjectAdd", param, true));
									}
									
									// Notify moving player of new player in AOI
									ArrayList param = new ArrayList();
									param.add((PlayerCharacter) objects.get(newPlayers.get(i1)));
									clients.get(pc.name).addInput(new Signal("ObjectAdd", param, true));
								}
								
								/*
								 * 
								 * PROXIES
								 * 
								 * 
								 */
								
				
								
								removeItems = new ArrayList<Integer>();
								
								// If player already knows the proxy is in his AOI, remove proxy.
								// If not add proxy to players AOIObjects list.
								for(int j = 0; j < newProxies.size(); j++) {
									if(pc.AOIObjects.contains(newProxies.get(j))) {
										//System.out.println("PC" + pc.objectID + "  Obj " + newProxies.get(j) + " removed (already exists)");
										removeItems.add(j);	
									} else {
										pc.AOIObjects.add(newProxies.get(j));
										//System.out.println("PC" + pc.objectID + " Obj " + newProxies.get(j) + " added");
									}
								}
								skew = 0;
								for(int j = 0; j < removeItems.size(); j++) {
									newProxies.remove((int) removeItems.get(j)-skew);
									skew++;
								}
								
								for(int i1 = 0; i1 < newProxies.size(); i1++) {
									// Notify moving player of new proxy in AOI
									ArrayList param = new ArrayList();
									param.add((PlayerCharacter) proxiesIn.get(newProxies.get(i1)));

									clients.get(pc.name).addInput(new Signal("ObjectAdd", param, true));
								}
								
							}
							pc.lastStateCheck = currentTime;
						}
						
						t = new SDLTimer(20, new Signal("CheckStates", null, true));
						t.start();
						timers.add(t);
						panel.repaint();
					}
					if(signal.equals("ClientLeftZone")) {
						System.out.println(now.getTime() + ": ZoneModel (" + Zone.ZoneID + ") Client Left Zone!");
						PlayerCharacter pc = (PlayerCharacter) s.parameters.get(0);
						ZMClientConnection zmc = this.clients.get(pc.name);
						
						ArrayList param = new ArrayList();
						param.add(pc);
						this.gsc.addInput(new Signal("ClientHasNewMaster", param, true));
						
						param = new ArrayList();
						Zone newZ = (Zone) s.parameters.get(1);
						param.add(adjacentZoneMasters.get(newZ.ZoneID).getRemoteIP());
						param.add(zmc.challengeAnswer);
						param.add(4510+newZ.ZoneID);
						zmc.addInput(new Signal("NewZoneMaster", param, true));
						
						zmc.running = false;
						this.clients.remove(pc.name);
						this.characters.remove(pc.name);
						this.objects.remove(pc.objectID);
					}
				}
			}
		}
		System.out.println("ZoneModel stopping.....");
		window.dispose();
		td.running = false;
		try {
			zmzmc.server.close();
		} catch (IOException e) {
			
		}
		zmzmc.running = false;
		
		Set<String> clients = this.clients.keySet();
		Iterator<String> clIter = clients.iterator();
		while(clIter.hasNext()) {
			String cnm = clIter.next();
			(this.clients.get(cnm)).running = false;
		}
		Set<Integer> adjacentZoneMasters = this.adjacentZoneMasters.keySet();
		Iterator<Integer> azmIter = adjacentZoneMasters.iterator();
		while(azmIter.hasNext()) {
			Integer azmid = azmIter.next();
			(this.adjacentZoneMasters.get(azmid)).running = false;
		}
	}

	/*
	 * Returns a list of all the players in the AOI (Area Of Interrest) of a gameobject.
	 * 
	 * @param p		GameObject thats AOI defines the area
	 * @return		GameObjectList containing all the players in the AOI
	 */
	private GameObjectList findPlayersInAOI(GameObject p) {
		GameObjectList ret = new GameObjectList();
		Set<String> keys = this.characters.keySet();
		
		Iterator<String> i = keys.iterator();
		while(i.hasNext()) {
			String key = i.next();
			PlayerCharacter pc = this.characters.get(key);
			double x = Math.abs((double)pc.posX - (double)p.posX);
			double y = Math.abs((double)pc.posY - (double)p.posY);
			double z = Math.abs((double)pc.posZ - (double)p.posZ);
			double d = Math.sqrt(x*x + y*y + z*z);
			
			if(d <= 15) ret.add(pc);
		}
		
		return ret;
	}
	
	/*
	 * Returns a list of all the proxies in the AOI (Area Of Interrest) of a gameobject.
	 * 
	 * @param p		GameObject thats AOI defines the area
	 * @return		GameObjectList containing all the proxies in the AOI
	 */
	private GameObjectList findProxiesInAOI(GameObject p) {
		GameObjectList ret = new GameObjectList();
		Set<Integer> keys = this.proxiesIn.keySet();
		
		Iterator<Integer> i = keys.iterator();
		while(i.hasNext()) {
			int key = i.next();
			PlayerCharacter pc = (PlayerCharacter) this.proxiesIn.get(key);
			double x = Math.abs((double)pc.posX - (double)p.posX);
			double y = Math.abs((double)pc.posY - (double)p.posY);
			double z = Math.abs((double)pc.posZ - (double)p.posZ);
			double d = Math.sqrt(x*x + y*y + z*z);
			
			if(d <= 15) ret.add(pc);
		}
		
		return ret;
	}
	
	/*
	 * Adds a signal to the back of the signal queue.
	 * 
	 * @param input The signal to be added
	 */
	public void addInput(Signal input) {
		synchronized(this.input) {
			this.input.add(input);
		}
	}
	
	/*
	 * Class showing the zone map
	 */
	public class DrawingPanel extends JPanel {
		ZoneModel zm;
		DrawingPanel (ZoneModel zm) {
		    setBackground (Color.WHITE);
		    this.zm = zm;
		}

		public void paintComponent(Graphics g)   {
		   super.paintComponent(g);
		   g.setColor (Color.BLACK);
		   int dy = getSize ().height;
		   int dx = getSize ().width;
		   
			Set<String> keys = zm.characters.keySet();
			
			Iterator<String> i = keys.iterator();
			while(i.hasNext()) {
				String key = i.next();
				PlayerCharacter pc = zm.characters.get(key);
				
				g.fillOval (((int)pc.posX-zm.Zone.x)*4-2, ((int)pc.posZ-zm.Zone.y)*4-2, 4, 4);
			}
		}
	}
	
	/*
	 * Parses an XML string to a list of GameObjects
	 * 
	 * @param xml	XML string
	 * @return		GameObjectList containing the objects
	 */
	public GameObjectList parseXMLString(String xml) {
		GameObjectList gol = new GameObjectList();
		
		try {
			DOMParser parser = new DOMParser();
			parser.parse(new InputSource(new java.io.StringReader(xml)));
			Document doc = parser.getDocument();

			doc.getDocumentElement().normalize();
			
			NodeList nodeLst = doc.getElementsByTagName("GameObject");
			for (int s = 0; s < nodeLst.getLength(); s++) {
				Node fstNode = nodeLst.item(s);
				if (fstNode.getNodeType() == Node.ELEMENT_NODE) {			  
					Element fstElmnt = (Element) fstNode;
					
					NodeList oidElmntLst = fstElmnt.getElementsByTagName("objectID");
					Element oidElmnt = (Element) oidElmntLst.item(0);
					NodeList oid = oidElmnt.getChildNodes();
					int oid_ = Integer.parseInt(((Node) oid.item(0)).getNodeValue());
					
					NodeList typeElmntLst = fstElmnt.getElementsByTagName("type");
					Element typeElmnt = (Element) typeElmntLst.item(0);
					NodeList type = typeElmnt.getChildNodes();
					String type_ = ((Node) type.item(0)).getNodeValue();
					
					NodeList posXElmntLst = fstElmnt.getElementsByTagName("posX");
					Element posXElmnt = (Element) posXElmntLst.item(0);
					NodeList posX = posXElmnt.getChildNodes();
					double posX_ = Double.parseDouble(((Node) posX.item(0)).getNodeValue());
					
					NodeList posYElmntLst = fstElmnt.getElementsByTagName("posY");
					Element posYElmnt = (Element) posYElmntLst.item(0);
					NodeList posY = posYElmnt.getChildNodes();
					double posY_ = Double.parseDouble(((Node) posY.item(0)).getNodeValue());
					
					NodeList posZElmntLst = fstElmnt.getElementsByTagName("posZ");
					Element posZElmnt = (Element) posZElmntLst.item(0);
					NodeList posZ = posZElmnt.getChildNodes();
					double posZ_ = Double.parseDouble(((Node) posZ.item(0)).getNodeValue());
					
					NodeList rotXElmntLst = fstElmnt.getElementsByTagName("rotX");
					Element rotXElmnt = (Element) rotXElmntLst.item(0);
					NodeList rotX = rotXElmnt.getChildNodes();
					double rotX_ = Double.parseDouble(((Node) rotX.item(0)).getNodeValue());
					
					NodeList rotYElmntLst = fstElmnt.getElementsByTagName("rotY");
					Element rotYElmnt = (Element) rotYElmntLst.item(0);
					NodeList rotY = rotYElmnt.getChildNodes();
					double rotY_ = Double.parseDouble(((Node) rotY.item(0)).getNodeValue());
					
					NodeList rotZElmntLst = fstElmnt.getElementsByTagName("rotZ");
					Element rotZElmnt = (Element) rotZElmntLst.item(0);
					NodeList rotZ = rotZElmnt.getChildNodes();
					double rotZ_ = Double.parseDouble(((Node) rotZ.item(0)).getNodeValue());
					
					GameObject go = new GameObject(oid_, type_, posX_, posY_, posZ_, rotX_, rotY_, rotZ_);
					gol.add(go);
				}
			}
			
			nodeLst = doc.getElementsByTagName("PlayerCharacter");
			for (int s = 0; s < nodeLst.getLength(); s++) {
				Node fstNode = nodeLst.item(s);
				if (fstNode.getNodeType() == Node.ELEMENT_NODE) {			  
					Element fstElmnt = (Element) fstNode;
					
					NodeList oidElmntLst = fstElmnt.getElementsByTagName("objectID");
					Element oidElmnt = (Element) oidElmntLst.item(0);
					NodeList oid = oidElmnt.getChildNodes();
					int oid_ = Integer.parseInt(((Node) oid.item(0)).getNodeValue());
					
					NodeList posXElmntLst = fstElmnt.getElementsByTagName("posX");
					Element posXElmnt = (Element) posXElmntLst.item(0);
					NodeList posX = posXElmnt.getChildNodes();
					double posX_ = Double.parseDouble(((Node) posX.item(0)).getNodeValue());
					
					NodeList posYElmntLst = fstElmnt.getElementsByTagName("posY");
					Element posYElmnt = (Element) posYElmntLst.item(0);
					NodeList posY = posYElmnt.getChildNodes();
					double posY_ = Double.parseDouble(((Node) posY.item(0)).getNodeValue());
					
					NodeList posZElmntLst = fstElmnt.getElementsByTagName("posZ");
					Element posZElmnt = (Element) posZElmntLst.item(0);
					NodeList posZ = posZElmnt.getChildNodes();
					double posZ_ = Double.parseDouble(((Node) posZ.item(0)).getNodeValue());
					
					NodeList rotXElmntLst = fstElmnt.getElementsByTagName("rotX");
					Element rotXElmnt = (Element) rotXElmntLst.item(0);
					NodeList rotX = rotXElmnt.getChildNodes();
					double rotX_ = Double.parseDouble(((Node) rotX.item(0)).getNodeValue());
					
					NodeList rotYElmntLst = fstElmnt.getElementsByTagName("rotY");
					Element rotYElmnt = (Element) rotYElmntLst.item(0);
					NodeList rotY = rotYElmnt.getChildNodes();
					double rotY_ = Double.parseDouble(((Node) rotY.item(0)).getNodeValue());
					
					NodeList rotZElmntLst = fstElmnt.getElementsByTagName("rotZ");
					Element rotZElmnt = (Element) rotZElmntLst.item(0);
					NodeList rotZ = rotZElmnt.getChildNodes();
					double rotZ_ = Double.parseDouble(((Node) rotZ.item(0)).getNodeValue());
					
					NodeList nameElmntLst = fstElmnt.getElementsByTagName("name");
					Element nameElmnt = (Element) nameElmntLst.item(0);
					NodeList name = nameElmnt.getChildNodes();
					String name_ = ((Node) name.item(0)).getNodeValue();
					
					NodeList movingElmntLst = fstElmnt.getElementsByTagName("moving");
					Element movingElmnt = (Element) movingElmntLst.item(0);
					NodeList moving = movingElmnt.getChildNodes();
					boolean moving_ = Boolean.parseBoolean(((Node) moving.item(0)).getNodeValue());
					
					PlayerCharacter go = new PlayerCharacter(oid_, posX_, posY_, posZ_, rotX_, rotY_, rotZ_, name_);
					go.moving = moving_;
					gol.add(go);
				}
			}
			
		} catch (Exception e) {
			e.printStackTrace();
		}
		return gol;
	}
	
	/*
	 * Updates all proxies in the zone (movement and AOI lists)
	 */
	public void checkProxies() {
		Set<Integer> keys = this.proxiesIn.keySet();
		
		Iterator<Integer> i = keys.iterator();
		while(i.hasNext()) {
			int key = i.next();
			PlayerCharacter pc = (PlayerCharacter) this.proxiesIn.get(key);

			Date now = new Date();
			long currentTime = now.getTime();
			
			if(pc.moving) {
				if(pc.lastStateCheck > 0) {
					long deltaTime = currentTime - pc.lastStateCheck;
					
					pc.posX += ((double)deltaTime/1000)*(1)*Math.sin(pc.rotY);
					pc.posZ += ((double)deltaTime/1000)*(-1)*Math.cos(pc.rotY);
					
				}
				
				GameObjectList affectedPlayers = this.findPlayersInAOI(pc);
				ArrayList<Integer> newPlayers = new ArrayList<Integer>();
				for(int j = 0; j < affectedPlayers.GameObjects.size(); j++) {
					newPlayers.add(affectedPlayers.GameObjects.get(j).objectID);
					//System.out.println("PC" + pc.objectID + "  In AOI: " + affectedPlayers.GameObjects.get(j).objectID);
				}
				
				ArrayList<Integer> removeItems = new ArrayList<Integer>();
				
				for(int j = 0; j < pc.AOIObjects.size(); j++) {
					if(!newPlayers.contains(pc.AOIObjects.get(j))) {
						//System.out.println("PC" + pc.objectID + "  Obj " + pc.AOIObjects.get(j) + " removed (inexistent)");
						removeItems.add(j);
					}
				}
				int skew = 0;
				for(int j = 0; j < removeItems.size(); j++) {
					pc.AOIObjects.remove((int)removeItems.get(j)-skew);
					skew++;
				}
				
				removeItems = new ArrayList<Integer>();
				
				for(int j = 0; j < newPlayers.size(); j++) {
					if(pc.AOIObjects.contains(newPlayers.get(j))) {
						//System.out.println("PC" + pc.objectID + "  Obj " + newPlayers.get(j) + " removed (already exists)");
						removeItems.add(j);	
					} else {
						pc.AOIObjects.add(newPlayers.get(j));
						//System.out.println("PC" + pc.objectID + " Obj " + newPlayers.get(j) + " added");
					}
				}
				skew = 0;
				for(int j = 0; j < removeItems.size(); j++) {
					newPlayers.remove((int) removeItems.get(j)-skew);
					skew++;
				}
				
				for(int i1 = 0; i1 < newPlayers.size(); i1++) {
					ZMClientConnection c = clients.get( ((PlayerCharacter) (objects.get(newPlayers.get(i1)))).name );
					
					// Notify other player of moving proxy entering AOI
					if(!((PlayerCharacter) objects.get(newPlayers.get(i1))).name.equals(pc.name)) {
						ArrayList param = new ArrayList();
						param.add((PlayerCharacter) pc);
						System.out.println("ZoneModel(" + Zone.ZoneID + ") notifying player (" + ((PlayerCharacter) (objects.get(newPlayers.get(i1)))).name + ") of proxy (" + pc.name + ")");
						c.addInput(new Signal("ObjectAdd", param, true));
					}
				}
				
				pc.lastStateCheck = currentTime;
			}
		}
	}
}
